"""
Mixins to handle common situations
"""
import os
import sys
import json
import csv
import copy
import shutil
import logging
import io
import hashlib
import tempfile
import traceback
from importlib import import_module
from inspect import signature
import multiprocessing as mp
import numpy as np
import pandas as pd
from inspect import getframeinfo, stack
from functools import partial

import s3fs
import gcsfs

from ..lib.misc import get_checksum, get_file_size
from ..lib.integration import send_html_email
from ..datasets import TransformDoodle
from .frames import PandasDataFrame

logger = logging.getLogger("app")

__all__ = [
    "S3Mixin",
    "GCSMixin",
    "PandasMixin",
    "ParallelMixin",
    "AWSMixin",
    "FilesMixin",
    "CheckpointMixin",
    "EmailMixin",
    "DatasetSpecMixin",
    "DoodleMixin"
]


class CloudMixin(object):
    """
    Mixin baseclass for S3 and GCSFS.
    """

    def _cloud_get_fshandle(self, fshandle=None, attr="s3"):
        """ """
        if fshandle is None:
            fshandle = getattr(self, attr, None)
        if fshandle is None:
            logger.error(
                "{} handle is missing".format(attr),
                extra=self.config.get_extra(
                    {
                        "transform": self.name,
                    }
                ),
            )
            raise Exception("Invalid {} handle".format(attr))

        return fshandle

    def _cloud_list_files(self, fshandle, path, include=None):

        files = []
        try:
            files = fshandle.glob(path)
        except:
            logger.exception(
                "Unable to access path",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Unable to access path")

        # => Split the files
        if include is None:
            include = lambda x: True

        if not callable(include):
            raise Exception("Include parameter should be a function")

        included_files = []
        excluded_files = []
        for f in files:
            if include(f):
                included_files.append(f)
            else:
                excluded_files.append(f)

        return included_files, excluded_files

    def _cloud_list_directories(self, fshandle, path, normalizer=None):
        files = []
        try:
            files = fshandle.ls(path, detail=True)
        except:
            logger.exception(
                "Unable to access path",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Unable to access path")

        dirs = []
        excluded_files = []

        for f in files:

            # Normalize
            if normalizer is not None and callable(normalizer):
                f = normalizer(f)

            if f["StorageClass"] == "DIRECTORY":
                dirs.append(f["Key"])
            else:
                excluded_files.append(f["Key"])

        return dirs, excluded_files

    def _cloud_open_file(self, fshandle, path):

        try:
            fd = fshandle.open(path, "rb")
        except:
            logger.exception(
                "Unable to read file",
                extra=self.config.get_extra(
                    {"transform": self.name, "data": path}
                ),
            )
            raise Exception("Unable to read blockstore file")

        return fd

    def _cloud_read_file(self, fshandle, path):

        try:
            content = fshandle.open(path, "rb").read()
        except:
            logger.exception(
                "Unable to read file",
                extra=self.config.get_extra(
                    {"transform": self.name, "data": path}
                ),
            )
            raise Exception("Unable to read blockstore file")

        return content

    def _cloud_write_file(self, fshandle, path, content):

        try:

            if isinstance(content, str):
                content = content.encode('utf-8')

            metadata = {
                "path": path,
                "sha256": hashlib.sha256(content).hexdigest(),
                "size": len(content)
            }

            with fshandle.open(path, "wb") as fd:
                fd.write(content)

        except:
            logger.exception(
                "Unable to write file",
                extra=self.config.get_extra(
                    {"transform": self.name, "data": path}
                ),
            )
            raise Exception("Unable to read blockstore file")

        return metadata

    def _cloud_write_frame(self, fshandle, path, df):
        """
        Write a dataframe
        """

        buf = io.StringIO()
        df.to_csv(buf, index=False, quoting=csv.QUOTE_NONNUMERIC)
        content = buf.getvalue()
        del buf

        metadata = self._cloud_write_file(fshandle,
                                          path=path,
                                          content=content)
        metadata.update({
            'rows': df.shape[0],
            'columns': df.shape[1],
        })

        return metadata


class S3Mixin(CloudMixin):
    """
    AWS S3 functions usable by any module
    """

    def get_s3_handle(self, cred=None, client_kwargs=None):
        """
        Backward compatability. Calls s3_init_fshandle
        """
        return self.s3_init_fshandle(cred, client_kwargs=client_kwargs)

    def s3_init_fshandle(self, cred=None, client_kwargs=None):
        """
        Open a S3 connection. Takes credentials as an explicit
        argument or can pickup from 'aws_cred' attribute of self.

        Args:

          cred (object): Optional Credentials (a dictionary with 'secret_key' and 'access_key' parameters)

        Returns:

          object: S3FS handle

        """

        if cred is None:
            if hasattr(self, "aws_cred"):
                cred = self.aws_cred

        fail = True
        msg = ""
        if cred is None:
            msg += "\ncred is None"

        if not isinstance(cred, dict):
            msg += "\naws credentials is not a dictionary"
        else:
            if "access_key" not in cred:
                msg += "\naccess_key missing in credentials"
            elif "secret_key" not in cred:
                msg += "\nsecret_key missing in credentials"
            else:
                fail = False

        if fail:
            logger.error(
                "Invalid aws credentials",
                extra=self.config.get_extra({"transform": self.name, "data": msg}),
            )
            raise Exception("Could not find the credentials")

        # Insert region if required.
        if 'region' in cred:
            if client_kwargs is None:
                client_kwargs = {}
            client_kwargs['region_name'] = cred.get('region')

        return s3fs.S3FileSystem(
            anon=False,
            key=cred.get("access_key",None),
            secret=cred.get("secret_key",None),
            token=cred.get('token', None),
            use_listings_cache=False,
            client_kwargs=client_kwargs,
        )

    def _s3_get_fshandle(self, s3):
        """
        Get S3 handle.

        Args:

           s3 (object): s3fs instance obtained from get_s3_handle
                        If not specified, 's3' object attribute used

        Returns:

          fileystem handle
        """

        return self._cloud_get_fshandle(fshandle=s3, attr='s3')

    def s3_list_files(self, path, s3=None, include=None):
        """
        List files specified by a glob (path)

        Args:

           s3 (object): s3fs instance obtained from get_s3_handle
                        If not specified, 's3' object attribute used
           path (str): Glob of files
           include (method): Function that is called for every file to determine whether to include or not

        Returns:

           tuple: List of included and excluded files

           first value is the list of included and the second is the list of excluded

        """

        s3 = self._s3_get_fshandle(s3=s3)
        return self._cloud_list_files(
            fshandle=s3, path=path, include=include
        )

    def s3_list_directories(self, path, s3=None):
        """
        List files specified by a glob (path)

        Args:

           s3 (object): s3fs instance obtained from get_s3_handle
                        If not specified, 's3' object attribute used
           path (str): Glob of files

        Returns:

           tuple: List of directories and non directories

           first value is the list of directories and the second is the list of  non directories

        """

        s3 = self._s3_get_fshandle(s3)
        return self._cloud_list_directories(fshandle=s3, path=path)

    def s3_open_file(self, path, s3=None):
        """
        Open a S3 file and return a file descriptor

        Args:

          s3 (object): s3fs instance obtained from get_s3_handle
                       If not specified, 's3' object attribute used
          path (str): Path of the file that must be opened

        Returns:

          object: File descriptor

        """
        s3 = self._s3_get_fshandle(s3)
        return self._cloud_open_file(fshandle=s3, path=path)

    def s3_read_file(self, path, s3=None):
        """
        Read the contents of a S3 object

        Args:

          s3 (object): s3fs instance obtained from get_s3_handle
                       If not specified, 's3' object attribute used
          path (str): Path of the file that must be opened

        Returns:

          str: content of the path

        """
        s3 = self._s3_get_fshandle(s3)
        return self._cloud_read_file(fshandle=s3, path=path)

    def s3_write_file(self, path, content, s3=None):
        """
        Write the contents to a S3 object

        Args:

          s3 (object): s3fs instance obtained from get_s3_handle
                       If not specified, 's3' object attribute used
          path (str): Path of the file that must be opened
          content (bytes): Content of the file

        Returns:

          metadata (dict): Dictionary having file metadata

        """

        s3 = self._s3_get_fshandle(s3)
        return self._cloud_write_file(
            fshandle=s3, path=path, content=content
        )

    def s3_write_frame(self, path, df, s3=None):
        """
        Write the contents to a S3 object

        Args:

          s3 (object): s3fs instance obtained from get_s3_handle
                       If not specified, 's3' object attribute used
          path (str): Path of the file that must be opened
          df (dataframe): Dataframe to write

        Returns:

          metadata (dict): Dictionary having file metadata

        """
        s3 = self._s3_get_fshandle(s3)
        return self._cloud_write_frame(
            fshandle=s3, path=path, df=df
        )


    def s3_cache_files(self, files, localdir, s3=None):
        """
        Cache files from S3 into local directory

        Args:
          files (list): List of paths
          localdir (str): Path to cache the file
          s3 (object): s3fs instance obtained from get_s3_handle
                       If not specified, 's3' object attribute used
        Returns:
          list: list of updated files
        """

        assert hasattr(self, "get_cache_dir")

        s3 = self._s3_get_fshandle(s3)
        metadatafile = os.path.join(localdir, "__cache_metadata__.json")
        try:
            metadata = json.load(open(metadatafile))
        except:
            metadata = {}

        refreshed = []

        for f in files:

            basename = os.path.basename(f)
            localfile = os.path.join(localdir, basename)

            # First collect metadata. f doesnt have bucket name
            info = s3.info(f)

            # Fix the format
            info["LastModified"] = info["LastModified"].isoformat()

            # Collect the filename on local disk
            if os.path.exists(localfile):
                statinfo = os.stat(localfile)
                size = statinfo.st_size
            else:
                size = -1

            # Add a md5 sum check
            #
            if (
                (basename in metadata)
                and (metadata[basename]["ETag"] == info["ETag"])
                and (size == info["Size"])
            ):
                continue

            # Receive the file...
            metadata[basename] = info
            refreshed.append(basename)

            # Get the file content
            fd = self.s3.open(fullpath, "rb")
            content = fd.read()
            fd.close()

            # Write it locally
            with open(localfile, "wb") as fd:
                fd.write(content)

            # Save the state as well
            with open(metadatafile, "w") as fd:
                fd.write(json.dumps(metadata, indent=4))

            logger.debug(
                "Cached file {}".format(basename),
                extra=self.config.get_extra(
                    {
                        "transform": self.name,
                        "data": "{} => {}".format(basename, localfile),
                    }
                ),
            )

        return refreshed


class GCSMixin(CloudMixin):
    """
    GCS Blockstore functions usable by any module
    """

    def metadata_normalizer(self, detail):
        """
        Modify the GCP Metadata to be consistent with S3 metadata

        """
        mapping = {
            # S3 Key: GCP Key
            "StorageClass": "storageClass",
            "ETag": "etag",
            "Key": "path",
            "Size": "size",
            # This is a string. Last modified is a datetime object
            "LastModified": "updated",
        }
        for k, v in mapping.items():
            detail[k] = detail[v]

        return detail

    def get_gcs_handle(self, cred=None):
        """
        Backward compatability. Calls gcs_init_fshandle
        """
        return self.gcs_init_fshandle(cred)

    def gcs_init_handle(self, cred=None):
        """
        Open a GCS connection. Takes credentials as an explicit
        argument or can pickup from 'gcs_cred' attribute of self.

        Args:

          cred (object): Credentials (a dictionary with 'keyfile')

        Returns:

          object: GCSFS handle

        """

        if cred is None:
            if hasattr(self, "gcs_cred"):
                cred = self.gcs_cred

        fail = True
        msg = ""
        if cred is None:
            msg += "\ncred is None"

        if not isinstance(cred, dict):
            msg += "\ngcs credentials is not a dictionary"
        else:
            if "keyfile" not in cred:
                msg += "\nkeyfile missing in credentials"
            else:
                fail = False

        if fail:
            logger.error(
                "Invalid GCS credentials",
                extra=self.config.get_extra({"transform": self.name, "data": msg}),
            )
            raise Exception("Could not find the credentials")

        return gcsfs.GCSFileSystem(token=cred["keyfile"])

    def _gcs_get_fshandle(self, gcs):
        """
        Get GCS handle.

        Args:

           gcs (object): GCSFS instance obtained from get_gcs_handle
                        If not specified, 'gcs' object attribute used

        Returns:

          fileystem handle

        """
        return self._cloud_get_fshandle(
            fshandle=gcs, attr="gcs"
        )

    def gcs_list_files(self, path, gcs=None, include=None):
        """
        List files specified by a glob (path)

        Args:

           path (str): Glob of files
           gcs (object): gcsfs instance obtained from get_gcs_handle
                        If not specified, 'gcs' object attribute used
           include (def): Function that is called for every file to determine whether to include or not

        Returns:

           tuple: List of included and excluded files

           first value is the list of included and the second is the list of excluded

        """
        gcs = self._gcs_get_fshandle(gcs)
        return self._cloud_list_files(
            fshandle=gcs, path=path, include=include
        )

    def gcs_list_directories(self, path, gcs=None):
        """
        List files specified by a glob (path)

        Args:

           path (str): Glob of files
           gcs (object): gcsfs instance obtained from get_gcs_handle
                        If not specified, 'gcs' object attribute used

        Returns:

           tuple: List of directories and non directories

           first value is the list of directories and the second is the list of  non directories

        """

        gcs = self._gcs_get_handle(gcs)
        return self._cloud_list_directories(
            fshandle=gcs, path=path, normalizer=self.metadata_normalizer
        )

    def gcs_open_file(self, path, gcs=None):
        """
        Open a GCS file and return a file descriptor

        Args:

          path (str): Path of the file that must be opened
          gcs (object): gcsfs instance obtained from get_gcs_handle
                       If not specified, 'gcs' object attribute used

        Returns:

          object: File descriptor

        """
        gcs = self._gcs_get_handle(gcs)
        return self._cloud_open_file(fshandle=gcs, path=path)

    def gcs_read_file(self, path, gcs=None, ):
        """
        Read the contents of a GCS object

        Args:

          path (str): Path of the file that must be opened
          gcs (object): gcsfs instance obtained from get_gcs_handle
                       If not specified, 'gcs' object attribute used

        Returns:

          str: content of the path

        """
        gcs = self._gcs_get_fshandle(gcs)
        return self._cloud_read_file(fshandle=gcs, path=path)

    def gcs_write_file(self, path, content, gcs=None):
        """
        Write the contents to a GCS object

        Args:

          path (str): Path of the file that must be opened
          content (bytes): Content of the file
          gcs (object): gcsfs instance obtained from get_gcs_handle
                       If not specified, 'gcs' object attribute used

        Returns:

           metadata (dict): Metadata of the output file

        """

        gcs = self._gcs_get_handle(gcs, fsonly=True)
        return self._cloud_write_file(
            fshandle=gcs, path=path, content=content
        )

    def gcs_write_frame(self, path, df, gcs=None):
        """
        Write the contents to a GCS object

        Args:

          path (str): Path of the file that must be opened
          df (dataframe): Pandas dataframe
          gcs (object): gcsfs instance obtained from get_gcs_handle
                       If not specified, 'gcs' object attribute used

        Returns:

           metadata (dict): Metadata of the output file

        """

        gcs = self._gcs_get_handle(gcs)
        return self._cloud_write_frame(
            fshandle=gcs, path=path, df=df
        )


class ParallelMixin(object):
    """
    Execute a function in parallel
    """

    def pexec_single(
        self,
        df,
        func,
        partitions=10,
        cores=1,
    ):
        """
        Take a single DF, split it and run funcs. Get either a combined
        dataframe or a list of computed outputs

        Args:

          df (obj): Pandas dataframe
          func (def): Function that must be called
          partitions (int): Number of dataframe partitions
          cores (int): Number of cores to run on

        Returns:
          object: list of func outputs or a combined dataframe

        """
        if df is None:
            return []

        assert callable(func)

        df_split = np.array_split(df, partitions)
        pool = mp.Pool(cores)
        _dfs = pool.map(func, df_split)
        pool.close()
        pool.join()

        return _dfs

    def pexec_multiple(self, dfs, func, cores=1):
        """
        Take a list of DFs, and run a func on each in parallel. Get
        either a combined dataframe or a list of computed outputs

        Args:

          dfs (obj): List of Pandas dataframes
          func (def): Function that must be called
          cores (int): Number of cores to run on

        Returns:
          object: list of func outputs or a combined dataframe

        """

        # Execute...
        pool = mp.Pool(cores)
        _dfs = pool.map(func, dfs)
        pool.close()
        pool.join()

        return _dfs


class PandasMixin(object):
    """
    Helper functions that load/store files
    """

    def pandas_read_file(self, name, path, params=None, test=False):
        """
        Read a pandas file

        Args:
           name (str): Label for the object. We use it to get the params
           path (str): Path of the local file that must be read
           params (dict): If explicitly passed then the function will not lookup
                          the pandas_params dictionary
           test (bool): Whether this is a test run. This gets a limited number of rows

        Returns:
           obj: a Pandas dataframe

        """

        if params is None:
            if not hasattr(self, "pandas_params"):
                raise Exception(
                    "Either params should be passed or it could be specified in pandas params"
                )
            if name not in self.pandas_params:
                raise Exception("{} params should be specified in pandas_params")
            params = self.pandas_params[name]

        if not os.path.exists(path):
            raise Exception("Missing file: {}".format(path))

        params = copy.copy(params)
        if test:
            params["nrows"] = 10000

        # Now load the file...
        df = pd.read_csv(path, **params)

        return df

    def pandas_read_s3file(self, name, path, params=None, test=False):
        """
        Read a pandas file but one that is in s3

        Args:
           name (str): Label for the object. We use it to get the params
           path (str): Path of the s3 file that must be read including the bucket name
           params (dict): If explicitly passed then the function will not lookup
                          the pandas_params dictionary
           test (bool): Whether this is a test run. This gets a limited number of rows

        Returns:
           obj: a Pandas dataframe

        """

        if params is None:
            if not hasattr(self, "pandas_params"):
                raise Exception(
                    "Either params should be passed or it could be specified in pandas params"
                )
            if name not in self.pandas_params:
                raise Exception("{} params should be specified in pandas_params")
            params = self.pandas_params[name]

        params = copy.copy(params)
        if test:
            params["nrows"] = 10000

        # Now load the file...
        with self.s3_open_file(path) as fd:
            df = pd.read_csv(fd, **params)

        return df


class AWSMixin(object):
    """
    AWS helper functions..
    """

    def get_aws_attributes(self, name):
        """
        Lookup credentials

        Credentials are stored in siteconf on a enrich deployment site. This
        function looks up the credentials, opens the s3 connections, and cleans
        the paths

        Args:
           name (str): Name for credentials

        Returns:
           dict: A dictionary with credentials, s3 hangle, bucket, and path

        """

        # First check what is in the s3
        cred = self.args[name]["credentials"]
        aws_cred = self.get_credentials(cred)

        # Now get the handle
        s3 = self.get_s3_handle(cred=aws_cred)

        # Get the rest of the parameters
        bucket = self.args[name]["bucket"]
        path = self.args[name]["path"]
        if path.endswith("/"):
            path = path[:-1]
        path = path

        return {"cred": aws_cred, "s3": s3, "bucket": bucket, "path": path}


class FilesMixin(object):
    """
    Helper functions for file operations
    """

    def file_split_apply(self, label, files, splitfunc, applyfunc):
        """
        A function to split a list based on a name obtained from a
        splitting function

        Args:
          label (str): name to be used for documentation
          files (list): List of strings/files/names
          splitfunc (def): func name -> partition-name
          applyfunc (def): func list -> result

        Returns:
          dict: A dictionary with an entry for each partition provided by splitfunc

        """

        # => Validate parameters
        checks = [
            ["label", isinstance(label, str)],
            ["files", isinstance(files, list)],
            ["splitfunc", callable(splitfunc)],
            ["applyfunc", callable(applyfunc)],
        ]

        for c in checks:
            if not c[1]:
                raise Exception("Invalid parameter: {}".format(c[0]))

        # First split the list into groups using the
        # split function
        groups = {}
        for f in files:
            name = splitfunc(f)
            if name in ["", None]:
                raise Exception("Invalid name resulted from split function")
            if name not in groups:
                groups[name] = []
            groups[name].append(f)

        result = {}
        keys = sorted(list(groups.keys()))
        for k in keys:
            res = applyfunc(k, groups[k])
            result[k] = res

            logger.debug(
                "Processed {}: {}".format(label, k),
                extra=self.config.get_extra(
                    {
                        "transform": self.name,
                    }
                ),
            )

        return result

    def file_preprocessed_read(self, root, load=True, create=True):
        """
        Load the metadata of preprocessed files

        Args:
          root (str): Directory of preprocessed files
          load (bool): if True load the preprocessed files into dataframes
          create (bool): if True create the root directory

        Returns:
          dict: Metadata dictionary

        """

        now = self.config.now()

        metadatafile = os.path.join(root, "__preprocess_metadata__.json")
        metadata = {"dfs": {}, "files": []}

        # Create the root if required
        try:
            if create:
                os.makedirs(root)
        except:
            pass

        if os.path.exists(metadatafile):
            try:
                metadata = json.load(open(metadatafile))
            except:
                logger.exception(
                    "Unable to read preprocessed metadata file",
                    extra=self.config.get_extra(
                        {
                            "transform": self.name,
                        }
                    ),
                )

        metadata["dfs"] = {}

        # Fix the root
        data_root = self.config.get_file("%(data_root)s")
        metadata["root"] = os.path.relpath(root, start=data_root)

        # Look through the
        if load:
            for detail in metadata["files"]:
                datafile = os.path.join(root, detail["filename"])
                params = detail["params"]
                name = detail["name"]
                if not os.path.exists(datafile):
                    raise Exception("Missing data file: {}".format(datafile))
                metadata["dfs"][name] = pd.read_csv(datafile, **params)
                detail["timestamp"] = now.isoformat()

        return metadata

    def file_preprocessed_write(self, metadata):
        """
        Write the preprocessed files

        Load the metadata of preprocessed files

        Args:
          metadata (dict): Loaded metadata with updates

        """

        # Create a temporary root
        root = metadata["root"]
        if not os.path.isabs(root):
            data_root = self.config.get_file("%(data_root)s")
            root = os.path.join(data_root, root)

        newroot = os.path.join(root, ".new")
        try:
            os.makedirs(newroot)
        except:
            pass
        dfs = metadata["dfs"]

        metadatafile = os.path.join(newroot, "__preprocess_metadata__.json")

        # => Write the files
        for detail in metadata["files"]:
            name = detail["name"]
            filename = detail["filename"]
            params = detail["params"]
            path = os.path.join(newroot, filename)
            try:
                os.makedirs(os.path.dirname(path))
            except:
                pass

            df = dfs[name]
            df.to_csv(path, index=False, **params)

        # Make a copy and dump the file
        metadatacopy = copy.copy(metadata)
        metadatacopy.pop("dfs")
        with open(metadatafile, "w") as fd:
            fd.write(json.dumps(metadatacopy, indent=4))

        data_root = self.config.get_file("%(data_root)s")
        metadata["root"] = os.path.relpath(root, start=data_root)
        metadata["dfs"] = dfs

        # Now move the contents to a old directory
        oldroot = os.path.join(metadata["root"], ".old")

        # Clear old data
        try:
            if os.path.exists(oldroot):
                shutil.rmtree(oldroot)
            os.makedirs(oldroot)
        except:
            pass

        # Move the current files to the old directory
        for f in os.listdir(root):
            if os.path.basename(f) in [".old", ".new"]:  # skip self.
                continue
            shutil.move(
                src=os.path.join(root, os.path.basename(f)),
                dst=os.path.join(oldroot, os.path.basename(f)),
            )

        # Move the current files to the old directory
        for f in os.listdir(newroot):
            shutil.move(
                src=os.path.join(newroot, os.path.basename(f)),
                dst=os.path.join(root, os.path.basename(f)),
            )

        # Remove the the new
        shutil.rmtree(newroot)

        return

    def file_preprocessed_update(self, metadata, name, filename, df, params, context):
        """
        Update the preprocessed metadata with new information

        Args:
          metadata (dict): Loaded metadata
          name (str): Label for the dataframe/filename being updated
          filename (str): Expected filename on the disk
          df (object): Dataframe that must be stored
          params (dict): Params to be used while saving the dataframe
          context (dict): Any additional information


        """

        assert metadata is not None
        assert "files" in metadata

        # => First check if the entry exists
        found = False
        for detail in metadata["files"]:
            if detail["name"] == name:
                detail.update(
                    {"filename": filename, "params": params, "context": context}
                )
                found = True

        if not found:
            metadata["files"].append(
                {
                    "name": name,
                    "filename": filename,
                    "params": params,
                    "context": context,
                }
            )

        # Update the df in all cases
        metadata["dfs"][name] = df

        return

    def file_preprocessed_get(self, metadata, name):
        """
        Lookup the preprocessed metadata for a given name

        Args:
          metadata (dict): Loaded metadata
          name (str): Label for the dataframe/filename to lookup

        Returns:
          tuple: A tuple of (loaded dataframe, details)
        """

        df = metadata["dfs"].get(name, None)
        for detail in metadata["files"]:
            if detail["name"] == name:
                return df, detail

        return None, None


class CheckpointMixin(object):
    def checkpoint(
        self,
        df,
        filename,
        output="pq",
        metafile=None,
        extra_metadata={},
        state=None,
        **kwargs,
    ):
        """
        Checkpoint a dataframe. Collect all stats

        Args:

          df (object): Dataframe
          filename (str): Output filename. Probably unresolved
          metafile (str): Metadata filename. If not specified filename +'.metadata.json'
          extra_metadata (dict): Any additional information to be logged
          state (object): Enrich pipeline state.
          kwargs (dict): Any extra parameters to be passed to the pandas

        Returns:

          dict: Metadata

        """

        if state is None:
            if not hasattr(self, "state"):
                logger.error(
                    "State is required as parameter or transform attribute",
                    extra={"transform": self.name},
                )
            else:
                state = self.state

        root = self.config.get_file("%(enrich_data_dir)s")
        filename = self.config.get_file(filename)
        relpath = os.path.relpath(filename, start=root)

        if metafile is None:
            metafile = filename + ".metadata.json"

        metarelpath = os.path.relpath(metafile, start=root)

        # Dump the dataframe
        if output == "pq":
            handler = partial(df.to_parquet, filename)
        elif output == "csv":
            handler = partial(df.to_csv, filename)
        elif output == "tsv":
            handler = partial(df.to_csv, filename, sep="\t")
        else:
            raise Exception("Unsupported checkpoint output format: {}".format(output))

        try:

            try:
                # Create a target directory..
                os.makedirs(os.path.dirname(filename))
            except:
                pass

            # First try storing the filename
            if output != "pq":
                handler(**kwargs)
                status = "success"
                error = ""
            else:
                # Parquet
                kwargs.pop("engine", "")
                status = "failure"
                for engine in ["fastparquet", "pyarrow"]:
                    try:
                        handler(engine=engine, **kwargs)
                        status = "success"
                        error = ""
                        break
                    except:
                        logger.exception(
                            "{}: Failed to write {}".format(
                                engine, os.path.basename(filename)
                            ),
                            extra={"transform": self.name},
                        )

                if status == "failure":
                    raise Exception("Both parquet engines failed to write")

        except Exception as e:
            logger.exception(
                "Failed to store checkpoint", extra={"transform": self.name}
            )
            raise

        metadata = self.get_default_metadata(state)

        # Collect caller information
        info = []
        for s in stack()[1:3]:
            caller = getframeinfo(s[0])
            info.append("%s:%d" % (caller.filename, caller.lineno))

        # Now if the output file exists, collect information
        # about the output file
        metadata.update(
            {
                "status": status,
                "error": error,
                "extra": extra_metadata,
                "stack": info,
                "filename": relpath,
                "size": get_file_size(filename),
                "metadata": metarelpath,
                "checksum": get_checksum(filename),
                "rows": df.shape[0],
            }
        )

        if "columns" not in metadata:
            framemgr = PandasDataFrame()
            columns = {}
            for c in list(df.columns):
                columns[c] = {
                    "touch": self.name,  # Who is introducing this column
                    "datatype": framemgr.get_generic_dtype(df, c),  # What is its type
                }
            metadata["columns"] = columns

        with open(metafile, "w") as fd:
            fd.write(json.dumps(metadata, indent=4))

        # Done
        return metadata


class EmailMixin(object):
    """
    Send email
    """

    def send_email_helper(
            self,
            sender,
            receivers,
            subject,
            body,
            attachment_prefix="",
            attachment_suffix="",
            frames={},
            reply_to=[],
            bcc=[],
            cred={}
    ):
        """
        Send email to a destination..
        """

        assert isinstance(frames, dict)

        try:
            tmpdir = tempfile.mkdtemp()

            filenames = []
            for name, df in frames.items():
                filename = os.path.join(
                    tmpdir, f"{attachment_prefix}{name}{attachment_suffix}.csv"
                )
                filenames.append(filename)
                df.to_csv(filename, index=False)

            send_html_email(
                body,
                sender=sender,
                reply_to=reply_to,
                bcc=bcc,
                receivers=receivers,
                subject=subject,
                attachments=filenames,
                cred=cred
            )

            logger.debug(
                "Email notification sent",
                extra={
                    "transform": self.name,
                    "data": "Receivers: " + json.dumps(receivers),
                },
            )
        except:
            logger.exception(
                "Unable to send email",
                extra={
                    "transform": self.name,
                },
            )

        if os.path.exists(tmpdir):
            try:
                shutil.rmtree(tmpdir)
            except:
                logger.exception(
                    "Unable to cleanup temporary directory",
                    extra={"transform": self.name, "data": tmpdir},
                )

class DatasetSpecMixin(object):
    """
    This is to be used with spec that identifies the
    datasets to be loaded
    """

    def get_input_dataset(self, name):

        tspec = self.transform_spec
        if 'input_datasets' not in tspec['spec']:
            raise Exception(f"input_datasets not present in the transform spec")

        datasets = tspec['spec']['input_datasets']
        dspec = None
        for d in datasets:
            if d['name'] == name:
                dspec = d
                break

        if dspec is None:
            raise Exception(f"Dataset not defined: {name}")

        source = d.get('source', d.get('lib', None))
        if source is None:
            raise Exception(f"Unable to find dataset definition. Spec is missing 'lib' element")

        try:
            mod = import_module(source)
            dataset = mod.get_dataset(name)
        except:
            # traceback.print_exc()
            logger.exception(f"Unable to load: {source}")
            return None

        return dataset

class DoodleMixin(object):
    """
    Access doodle to update metadata
    """

    def update_doodle(self, dataset, filename, action="read", state=None):
        """
        Update Doodle server metadata with access information
        """
        try:

            if state is None:
                if hasattr(self,'state'):
                    state = self.state
                else:
                    state = self.config.state

            if state is None:
                logger.error("State required for updating doodle",
                             extra={
                                 'transform': self.name
                             })
                return

            # => First check for doodle spec
            if (('doodle' not in self.args) or (isinstance(self.args['doodle'], str))):
                credname = self.args.get('doodle', 'doodle')
                self.args['doodle'] = self.get_credentials(credname)
            doodlecred = self.args['doodle']

            # Instantiate the doodle
            doodle = TransformDoodle(self,
                                     self.state,
                                     cred=doodlecred,
                                     action="read")

            # Now figure out the source
            spec = dataset.get_doodle_source(filename)

            # Search
            sources = doodle.search_sources(**spec)
            if len(sources) == 0:
                logger.error(f"Missing in Doodle: {spec['name']}:{spec['version']}",
                             extra={
                                 'transform': self.name
                             })
                return

            source = sources[0]

            # The metadata is automatically inserted
            doodle.update_source(source['id'], {})

            logger.debug(f"Updated Doodle: {spec['name']}:{spec['version']}",
                         extra={
                             'transform': self.name
                         })
        except:
            logger.exception("Unable to update doodle",
                             extra={
                                 'transform': self.name
                             })



import os
import sys
import json
import logging
import traceback

logger = logging.getLogger("app")


def should_skip(name):

    return name.startswith(".") or name == "__pycache__"


def load_customer_assets(context={}):
    """
    Go through the customer directories and add the assets from all
    installed customer modules.

    The ENRICH_CUSTOMERS looks like system path, and it is processed
    as such. It gives priority to the modules earlier in the path
    than later. Typically you create a run directory and prepend
    that run directory in front of the standard directory.
    """

    assets = {}

    # Check if the environment is correctly configured
    enrich_customers = context.get(
        "ENRICH_CUSTOMERS", os.environ.get("ENRICH_CUSTOMERS", None)
    )

    if enrich_customers is None:
        enrich_root = context.get("ENRICH_ROOT", os.environ.get("ENRICH_ROOT", None))
        if enrich_root is not None:
            enrich_customers = os.path.join(enrich_root, "customers")

    if enrich_customers is None:
        logger.error("Customer root is missing : {}".format(enrich_customers))
        return assets

    # => Collect all possible paths..
    customer_roots = {}
    for customer_root in enrich_customers.split(";"):
        if not os.path.exists(customer_root):
            continue
        for customer in os.listdir(customer_root):
            if customer not in customer_roots:
                customer_roots[customer] = os.path.join(customer_root, customer)

    # Look at even the current working directory...
    cwd = os.getcwd()
    for parent in ["..", "../.."]:
        root = os.path.abspath(os.path.join(cwd, parent))
        customer = os.path.basename(root)
        if customer in customer_roots:
            continue
        customer_roots[customer] = root

    # Now go through each of them...
    for customer, customer_dir in customer_roots.items():

        # customer : scribble
        # customer_dir : $ENRICH_ROOT/customers/scribble

        # parent : $ENRICH_ROOT/customers
        customer_root = os.path.dirname(customer_dir)

        if not os.path.isdir(customer_dir) or should_skip(customer):
            continue

        enrichfile = os.path.join(customer_dir, "enrich.json")
        if not os.path.exists(enrichfile):
            #logger.debug("Missing enrichfile. Skipping customer: {}".format(customer))
            continue

        try:
            enrich = json.load(open(enrichfile))
        except:
            logger.debug("Invalid enrichfile. Skipping customer: {}".format(customer))
            continue

        if not enrich.get("enable", True):
            #logger.debug("Disabled customer. Skipping customer: {}".format(customer))
            continue

        for usecase in os.listdir(customer_dir):
            """
            Each usecase within the customer organization
            """
            usecasedir = os.path.join(customer_dir, usecase)

            if not os.path.isdir(usecasedir) or should_skip(usecase):
                continue

            enrichfile = os.path.join(usecasedir, "enrich.json")

            if not os.path.exists(enrichfile):
                logger.debug(
                    "Missing enrichfile. Skipping usecase: {}:{}".format(
                        customer, usecase
                    )
                )
                continue

            try:
                enrich = json.load(open(enrichfile))
            except:
                logger.debug(
                    "Invalid enrichfile. Skipping usecase: {}:{}".format(
                        customer, usecase
                    )
                )

            if not enrich.get("enable", True):
                logger.debug(
                    "Disabled org. Skipping usecase: {}:{}".format(customer, usecase)
                )
                continue

            # => Check assets
            # Contrib
            #        /pkg
            #            /assets
            #        /assets

            for assetsdir in [
                os.path.join(usecasedir, "pkg", "assets"),
                os.path.join(usecasedir, "assets"),
            ]:
                if not os.path.exists(assetsdir):
                    continue

                for a in os.listdir(assetsdir):
                    if should_skip(a):
                        continue

                    for pkgdir in [
                        os.path.join(assetsdir, a, a),
                        os.path.join(assetsdir, a, "src", a),
                    ]:

                        if not os.path.exists(pkgdir):
                            continue

                        libdir = os.path.abspath(os.path.dirname(pkgdir))
                        if libdir not in sys.path:
                            sys.path.append(libdir)

                        assets[a] = {
                            "fullpath": os.path.join(libdir, a),
                            "relpath": os.path.relpath(
                                os.path.join(libdir, a), customer_root
                            ),
                            "organization": enrich["org"]["name"],
                            "usecase": usecase,
                            "customer": customer,
                        }

    return assets


def find_usecase(path):
    """
    Given any filepath, find the usecase corresponding to it.
    """
    try:

        # Look through the loaded usecases from settings.
        from django.conf import settings

        path = os.path.realpath(path)
        for usecase in settings.USECASES:
            usecase_root = os.path.realpath(usecase["root"])
            if os.path.commonprefix([path, usecase_root]) == usecase_root:
                return usecase
    except:
        logger.exception("Could not find the usecase for the path")

    return None



import os
import logging
import json
import traceback
import requests
import numpy as np
from datetime import datetime, timedelta

from enrichsdk.lib import get_credentials_by_name
from enrichsdk.utils import SafeEncoder, make_safely_encodable
from enrichsdk.datasets.generators import handlers as shared_generators

logger = logging.getLogger('app')

class Doodle(object):

    def __init__(self, cred, *args, **kwargs):
        """
        default
        """
        self.cred = cred
        self.generators = kwargs.pop('generators', {})
        self.purpose = None
        super().__init__(*args, **kwargs)

    def get_url(self):
        """
        Get base url
        """
        cred = self.cred
        if isinstance(cred, str):
            server = cred
        elif isinstance(cred, dict) and ('url' in cred):
            server = cred['url']
        elif isinstance(cred, dict) and ('uri' in cred):
            server = cred['uri']
        else:
            raise Exception("Unable to compute the url")

        server += "/docs"
        return server

    def access_server(self, path, params={}, data={}, method='get'):
        """
        Get/Post from doodle server
        """

        cred = self.cred
        if isinstance(cred, str):
            server = cred
        elif isinstance(cred, dict) and ('url' in cred):
            server = cred['url']
        elif isinstance(cred, dict) and ('uri' in cred):
            server = cred['uri']
        else:
            raise Exception("Invalid server details")

        if server.endswith("/"):
            server = server[:-1]

        if path.startswith("/"):
            path = path[1:]

        headers = {
            'Accept': 'application/json'
        }

        # Try all possible keys
        for k in ['key', 'api_key', 'apikey']:
            if k in cred:
                headers['X-API-Key'] = cred[k]
                break

        extra= { "verify": False }
        if 'basicauth' in cred:
            auth = (cred['basicauth']['username'], cred['basicauth']['password'])
            extra['auth'] = auth

        url = f"{server}/{path}"

        try:
            if method == 'post':
                response = requests.post(url, params=params, headers=headers, json=data, **extra)
            elif method == 'patch':
                response = requests.patch(url, params=params, headers=headers, json=data, **extra)
            elif method == 'get':
                response = requests.get(url, params=params, headers=headers, **extra)
            else:
                raise Exception(f"Unknown method: {method}")
        except Exception as e:
            logger.exception("Unable to access server")
            return 500, str(e)

        try:
            result = response.json()
        except:
            result = {
                'error': response.content.decode('utf-8')
            }


        return response.status_code, result


    def list_catalogs(self,
                      only_active=True,
                      offset: int = 0,
                      limit: int = 10,
                      order_by: str = None):
        """
        Search the catalogs
        """

        params = {
            'only_active': only_active,
            'offset': offset,
            'limit': limit,
            'order_by': order_by
        }

        status, result = self.access_server("/api/v1/catalogs",
                                            params=params)
        if status >= 300:
            raise Exception(f"Could not list catalogs")

        return result

    def get_catalog(self, catalog_id):
        """
        Get the details of one catalog
        """
        status, result = self.access_server(f"/api/v1/catalogs/{catalog_id}")

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)
        return result

    def search_catalogs(self,
                      only_active=True,
                      name: str = None,
                      version: str = None,
                      offset: int = 0,
                      limit: int = 10,
                      query: str = None,
                      modified_since: datetime = None,
                      modified_before: datetime = None,
                      order_by: str = None):
        """
        Search the catalogs
        """

        params = {
            'only_active': only_active,
            'name': name,
            'version': version,
            'offset': offset,
            'limit': limit,
            'query': query,
            'modified_since': modified_since,
            'modified_before': modified_before,
            'order_by': order_by
        }

        status, result = self.access_server(f"/api/v1/catalogs/search",
                                            params=params)

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result

    def find_catalog(self, name, version):
        """
        Find one catalog with a precise name and version
        """

        result = self.search_catalogs(name=name,
                                      version=version)

        if len(result) > 1:
            raise Exception(f"Multiple {len(result)} catalogs. Expecting one")

        if len(result) == 0:
            raise Exception(f"No catalog found {name}:{version}")

        return result[0]

    ##################################################
    # Lookup and update sources
    ##################################################
    def list_sources(self,
                     catalog_id=None,
                     only_active=True,
                     offset: int = 0,
                     limit: int = 1000,
                     order_by: str = None):
        """
        List available sources for a catalog
        """

        params = {
            'catalog_id': catalog_id,
            'only_active': only_active,
            'offset': offset,
            'limit': limit,
            'order_by': order_by
        }

        status, result = self.access_server("/api/v1/sources",
                                            params=params)

        if status >= 300:
            logger.error("Failed to list sources",
                         extra={
                             'data': str(result)
                         })
            raise Exception(f"Could not find sources for catalog {catalog_id}")

        return result

    def get_source(self, source_id):
        """
        Update the feature with latest dataset information
        """
        status, result = self.access_server(f"/api/v1/sources/{source_id}")

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)
        return result

    def search_sources(self,
                       only_active=True,
                       name: str = None,
                       version: str = None,
                       catalog_id: str = None,
                       offset: int = 0,
                       limit: int = 10,
                       query: str = None,
                       modified_since: datetime = None,
                       modified_before: datetime = None,
                       order_by: str = None):
        """
        Search the sources
        """

        params = {
            'only_active': only_active,
            'name': name,
            'version': version,
            'catalog_id': catalog_id,
            'offset': offset,
            'limit': limit,
            'query': query,
            'modified_since': modified_since,
            'modified_before': modified_before,
            'order_by': order_by
        }

        status, result = self.access_server(f"/api/v1/sources/search",
                                            params=params)

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result


    def update_source(self, source_id, details):
        """
        Update the catalog with latest dataset information
        """

        status, result = self.access_server(f"/api/v1/sources/{source_id}",
                                            data=details,
                                            method="post")

        if status >= 300:
            logger.error("Failed to update source",
                         extra={
                             'data': str(result)
                         })
            raise Exception(f"Could not update source {source_id}")

        return result

    def add_source(self, catalog_id, details):
        """
        Update the catalog with latest dataset information
        """

        details['catalog_id'] = catalog_id
        status, result = self.access_server("/api/v1/sources",
                                            data=details,
                                            method="post")

        if status >= 300:
            logger.error("Failed to update source",
                         extra={
                             'data': str(result)
                         })
            raise Exception(f"Could not add source to {catalog_id}")

        return result

    ##################################################
    # Lookup and update features
    ##################################################
    def list_features(self,
                      source_id=None,
                      catalog_id=None,
                      only_active=True,
                      offset: int = 0,
                      limit: int = 5000,
                      order_by: str = None):

        """
        List available features for a source
        """


        params = {
            'source_id': source_id,
            'catalog_id': catalog_id,
            'only_active': only_active,
            'offset': offset,
            'limit': limit,
            'order_by': order_by
        }
        status, result = self.access_server("/api/v1/features",
                                            params=params)

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result

    def get_feature(self, feature_id):
        """
        Update the feature with latest dataset information
        """
        status, result = self.access_server(f"/api/v1/features/{feature_id}")


        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result

    def search_features(self,
                        only_active=True,
                        name: str = None,
                        version: str = None,
                        catalog_id: str = None,
                        source_id: str = None,
                        offset: int = 0,
                        limit: int = 10,
                        query: str = None,
                        modified_since: datetime = None,
                        modified_before: datetime = None,
                        order_by: str = None):
        """
        Search the features
        """

        params = {
            'only_active': only_active,
            'name': name,
            'version': version,
            'catalog_id': catalog_id,
            'source_id': source_id,
            'offset': offset,
            'limit': limit,
            'query': query,
            'modified_since': modified_since,
            'modified_before': modified_before,
            'order_by': order_by
        }

        status, result = self.access_server(f"/api/v1/features/search",
                                            params=params)

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result

    def update_feature(self, feature_id, details):
        """
        Update the feature with latest dataset information
        """
        status, result = self.access_server(f"/api/v1/features/{feature_id}",
                                            data=details,
                                            method="post")

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)
        return result

    def add_feature(self, catalog_id, source_id, details):
        """
        Update the source with latest feature information
        """

        details['catalog_id'] = catalog_id
        details['source_id'] = source_id

        status, result = self.access_server("/api/v1/features",
                                            data=details,
                                            method="post")

        if status >= 300:
            error = json.dumps(result['detail'])
            raise Exception(error)

        return result

    ####################################################
    # Performance
    ####################################################
    def get_performance(self, pid):
        status, result = self.access_server("/api/v1/performance",
                                            params={
                                                'pid': str(pid)
                                            })
        if status >= 300:
            raise Exception(f"Could not obtain performance data")

        return result

    ####################################################
    # Read source
    ####################################################
    def compute_source_paths(self, source, start, end):
        """
        Read the source data from start to end dates...
        """

        """
        Sample matching specification
        {
            "generate": "generate_datetime_daily",
            "compare": "compare_datetime_pattern_match_range",
            "pattern": "%Y-%m-%d",
            "filename": "data.csv",
            "searchpaths": [
                "enrich-terrapay/backup/tip.terrapay.com/rawdata/queries/txn_value"
            ],
            "matchingpath": "enrich-terrapay/backup/tip.terrapay.com/rawdata/queries/txn_value/2022-11-07/data.csv"
        }
        """

        try:
            match = source['details']['match']
        except:
            raise Exception("Match section is missing")

        generator = match.get('generate', None)
        if generator is None:
            raise Exception("No generator specified")
        elif ((generator not in self.generators) and
              (generator not in shared_generators)):
            raise Exception(f"Unknown generator: {generator}")

        # Special case of a fixed path
        if generator == 'static_path':
            if 'matchingpath' in match:
                return [match['matchingpath']]
            else:
                raise Exception("matchingpath is missing")


        # Rest of the cases
        generator = self.generators.get(generator, shared_generators.get(generator))

        # Get all the subdirs
        subdirs = generator(match, start, end)
        root = match['searchpaths'][0]
        filename = match['filename']

        paths = []
        for subdir in subdirs:
            paths.append(os.path.join(root, subdir['name'], filename))

        return paths

    def get_source_paths(self,
                         start, end,
                         name=None,
                         version="v1",
                         source_id=None):
        """
        Find all the source paths
        """
        if ((name is None) and (source_id is None)):
            raise Exception("name or source_id must be specified")

        if name is not None:
            sources = self.search_sources(name=name, version=version)
            if len(sources) > 0:
                source = sources[0]
            else:
                source = self.get_source(source_id)
        else:
            source = self.get_source(source_id)

        if not source['active']:
            logger.error("Inactive source: {source['id']}")

        paths = self.compute_source_paths(source, start, end)

        return source, paths


class TransformDoodle(Doodle):
    """
    Extension to Doodle to allow passing of transform
    information to the doodle server
    """

    def __init__(self, transform, state, *args, **kwargs):
        """
        """
        self.transform = transform
        self.state = state
        self.action = kwargs.pop('action', 'read')

        super().__init__(*args, **kwargs)


    def insert_access_metadata(self, details):
        """
        """
        metadata = self.transform.get_default_metadata(self.state)
        pipeline = metadata['pipeline']

        data = {
            "name": pipeline['name'],
            "usecase": pipeline['usecase'],
            "nature": "pipeline",
            "action": self.action
        }
        key = "%(usecase)s:%(name)s:%(nature)s:%(action)s" % data

        if 'metadata' not in details:
            details['metadata'] = {}
        if 'access' not in details['metadata']:
            details['metadata']['access'] = {}
        details['metadata']['access'][key] = data

    def add_source(self, catalog_id, details):
        """
        Insert access metadata before posting
        """
        self.insert_access_metadata(details)
        return super().add_source(catalog_id, details)

    def update_source(self, source_id, details):
        """
        Insert access metadata before posting
        """
        self.insert_access_metadata(details)
        return super().update_source(source_id, details)

if __name__ == "__main__":

    cred = get_credentials_by_name('doodle')

    doodle = Doodle(cred)

    sources = doodle.list_sources()
    for source in sources:
        print("Processing", source['name'])
        try:
            source, paths = doodle.get_source_paths(start=datetime.today() + timedelta(days=-7),
                                                    end=datetime.today(),
                                                    name=source['name'],
                                                    version=source['version'])

            print(json.dumps(paths, indent=4))
        except Exception as e:
            print("Error!", e)



from uuid import uuid4

from fastapi.concurrency import run_in_threadpool

from .basegpt import BaseGPT
from .lib import *
from ..agents import LLMMultiLabelTextClassifier

logger = get_logger()


class MultiClassifierGPT(BaseGPT):

    def __init__(self):
        super().__init__()
        self.name = "MultiClassifierGPT"
        self.agentname = "multiclassifiergpt"
        self.platform = "openai"
        self.namespace = "multiclassifiergpt"

    async def get_llm_agent(self):
        agentname = "multiclassifiergpt"
        platform = "openai"
        survey_agent = LLMMultiLabelTextClassifier(name=agentname, platform=platform)

        logger.debug(f"Built agent for {agentname} and platform {platform}")

        return {
            "agent": survey_agent
        }

    async def qna_run(self, request_id):
        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Failure",
                         extra={
                             'request_id': "invalid",
                             "source": "service",
                         })
            cache[request_id] = {
                'status': "failure",
                "message": f"Invalid request id"
            }
            return

        # First get the params
        value = cache[request_id]

        try:

            params = value['params']
            user = params['user']
            dataset = params['dataset']
            query = params['query']

            stats['query_count'] += 1

            label = f"{user}_{dataset}"

            # First get the agent...
            logger.debug(f"Getting agent",
                         extra={
                             "source": "service",
                             "user": user,
                             "dataset": dataset,
                             "request_id": request_id,
                             'data': json.dumps(value, indent=4, cls=SafeEncoder)
                         })

            stats['datasets'][label] = {
                'loaded': datetime.now().replace(microsecond=0).isoformat(),
                'username': user,
                'agent_created': True,
                'agent_status': "Created",
                "query_count": 0,
                "query_success": 0,
                "query_failure": 0,
            }

            # Now run the query
            success, result = await run_in_threadpool(
                lambda: self.agent.label(query, label_spec=self.get_customer_profile_spec()))

            if success:
                json_result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))

                stats['query_success'] += 1
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_success'] += 1
                query_update_result(request_id, {
                    "status": "success",
                    "result": json_result
                })
            else:
                stats['query_failure'] += 1
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_failure'] += 1
                query_update_result(request_id, {
                    "status": "failure",
                    "result": 'Error'
                })
                logger.exception(f"Failed to run query",
                                 extra={
                                     "source": "service",
                                     'request_id': request_id,
                                     "user": params.get('user', "unknown"),
                                     "dataset": params.get('dataset', "unknown"),
                                 })

        except Exception as e:
            stats['query_failure'] += 1
            stats['datasets'][label]['query_count'] += 1
            stats['datasets'][label]['query_failure'] += 1
            logger.exception(f"Failed to run query",
                             extra={
                                 "source": "service",
                                 'request_id': request_id,
                                 "user": params.get('user', "unknown"),
                                 "dataset": params.get('dataset', "unknown"),
                             })
            query_update_result(request_id, {
                "status": "failure",
                "answer": f"Unable to construct the answer. Could be an internal agent error. See the agent log",
                "result": {
                    "text": str(e)
                }
            })

    def get_customer_profile_spec(self):
        default = {

            "persona": """You are a highly advanced, AI-enabled bot whose job is to take a question from a user and assign two labels to it.
    The first label is what kind of insight is being asked about.
    The second label is what kind of cohort is being asked about.""",

            "instructions": """The questions will be about the results of a survey across multiple respondent cohorts for multiple products.
    Note that sometimes a question can be about all the products in the survey without asking about population cohorts, and sometimes cohort information will be asked.
    The three types of population cohorts are: region, gender, age
    
    The following are metrics that could be mentioned in the question:
      - 'reach' a.k.a. 'line insights' a.k.a. 'le'
      - 'voc' a.k.a. 'sentiment score' a.k.a. 'ss'
    
    Here are the rules by which you must perform the insight label assignment:
      - If the question is asking about actionable insights, then assign insight=actionable_insights
      - If the question is asking specifically about best performing or worst performing products, then assign insight=actionable_insights
      - If the question is asking about product metrics with a specific cohort, then assign insight=le_insights
      - If the question is asking about the affinity of a population cohort to products, then assign insight=ss_insights
      - If the question is asking about product features, then assign insight=features_insights
      - If the question is asking about the audience, then assign insight=audience_insights
    
    And here are the rules by which you must perform the cohort label assignment:
      - If the question is asking about groups of respondents, then assign cohort=audience
      - If the question is asking about locations, then assign cohort=region
      - If the question is asking about gender, then assign cohort=gender
      - If the question is asking about age, then assign cohort=age
      - If the question is asking about best, or winners, or hits then assign cohort=best
      - If the question is asking about worst or losers then assign cohort=worst
      - If the question is asking about not liking something, then assign cohort=dislikes
      - If the question is asking about one metric being high and another being low, then assign cohort=mixed""",

            "synonymns": {
                "actionable insights": [
                    "survey summary",
                    "global summary",
                    "overall performance"
                ],
                "line efficiency": [
                    "le",
                    "reach",
                    "cannibalization",
                    "incremental reach",
                    "cumulative reach",
                ],
                "sentiment score": [
                    "ss",
                    "voc",
                    "voice of consumer",
                    "sentiment graph",
                    "consumer graph",
                    "consumer appetite",
                    "consumer sentiment",
                ],
            },

            "output_format": {
                "type": "json",
                "sample": '{"insight": insight, "cohort": cohort}'
            }
        }

        return default


app = FastAPI()

##############################################
# Common
##############################################
add_show_cache(app)
add_show_health(app)
policyspec = []
add_policy(app, policyspec)
multClassifierGpt = MultiClassifierGPT()
app.include_router(multClassifierGpt.router)



"""
Extractors
^^^^^^^^^^

Classes to support querying of datasets from database.

Implementation of this class for SQLite is in enrich-scribble assets.

"""


class ExtractorBase(object):
    """
    Baseclass for an engine to query datasets

    """

    def __init__(self, conf, *args, **kwargs):
        self.conf = conf

    def get_metadata(self):
        """
        Get metdata for the dataset specified
        """
        raise Exception("Implement in subclass")

    def get_collections_metadata(self, table):
        """
        Get metdata for a given table

        Args:
          table (str): Collection whose metadata is required
        """
        raise Exception("Implement in subclass")

    def get_attributes_metadata(self, filterargs):

        """
        Get attribute information for a given dataset

        Args:
          filterargs (dict): Collection and other information
             {
               'table': 'XYA'
             }

        """
        raise Exception("Implement in subclass")

    def query_by_spec(self, spec, filterargs):
        """
        Specify an abstract query in spec and scope
        in the filter args (table, output columns)

        Args:
          spec (dict): Specification of the query
          filterargs (dict): Has table and output columns
        """
        raise Exception("Implement in subclass")

    def query_raw(self, query):
        """
        Specify a raw to be executed. This reqires that
        the caller has an idea of what the dataset looks
        like and how to compute

        Args:
          query (object): Any object that the engine can accept
        """
        raise Exception("Implement in subclass")



__all__ = [
    "FileOperationsBase",
    "FileBasedQueryExecutorBase",
    "NotebookExecutorBase",
    "InMemoryQueryExecutorBase",
    "FeatureComputeBase",
    "MetricsBase",
    "AnomaliesBase",
    "SyntheticDataGeneratorBase",
    "DataObserverBase",
    "ChangePointDetectorBase",
    "TimeSeriesForecasterBase",
    "ClassifierBase",
    'DataSanitizerBase',
    'CustomEncoder'
]

from enrichsdk.utils import SafeEncoder as CustomEncoder

from .fileops import *
from .filebased_query_executor import *
from .inmemory_query_executor import *
from .notebook_executor import *
from .feature_compute import *
from .metrics import *
from .anomalies import *
from .changepoints import *
from .classifier import *
from .timeseries_forecaster import *
from .synthetic_data_generator import *
from .observability import *
from .data_quality import *
from .data_sanitizer import *



import openai
from enrichsdk.lib import get_credentials_by_name

__all__ = [
    'LLMTextGenerator'
]
class LLMTextGenerator(object):

    def __init__(self, cred, *args, **kwargs):
        """
        default
        """
        # vars
        self.available_models = {
            "completion": [
                    "text-davinci-003",
                    "text-curie-001",
                    "text-babbage-001",
                    "text-ada-001",
                    "code-davinci-002",
                    "code-cushman-001"
                ],
            "embedding": [
                    "text-embedding-ada-002"
                ]
        }
        self.default_code_model = "code-davinci-002"
        self.default_text_model = "text-davinci-003"
        self.default_embedding_model = "text-embedding-ada-002"

        # setup defaults
        # use a valid model
        model = kwargs.pop('model', "")
        available_models = []
        for task, models in self.available_models.items():
            available_models += models
        if model not in available_models:
            model = self.default_text_model
        self.model = model

        # setup the api key
        self.cred = cred
        self.api_key = self.get_api_key(self.cred)

        super().__init__(*args, **kwargs)

    def get_api_key(self, cred):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and 'apikey' in cred:
            api_key = cred['apikey']
        return api_key

    def set_model(self, task, model):
        """
        set the model to use for text completion
        """
        available_models = self.available_models.get(task, [])
        if model not in available_models:
            return False

        self.model = model
        return True

    def get_model(self):
        """
        get the model to use for text completion
        """
        return self.model

    def generate_text(self, **kwargs):
        """
        generate a text completion given a prompt
        """
        kwargs['model'] = self.default_text_model
        return self.generate_common(task='completion', **kwargs)

    def generate_code(self, **kwargs):
        """
        generate a code completion given a prompt
        """
        kwargs['model'] = self.default_code_model
        return self.generate_common(task='completion', **kwargs)

    def generate_embedding(self, **kwargs):
        """
        generate an embedding vector given some text
        """
        kwargs['model'] = self.default_embedding_model
        return self.generate_common(task='embedding', **kwargs)

    def generate_common(self, task, **kwargs):
        """
        generate a completion given a prompt
        """
        # default, override if successful
        result = {"success": False}

        # check if we have all the required params
        if self.api_key == None:
            result['msg'] = "No API key available"
            return result

        prompt = kwargs.get('prompt')
        if prompt == None:
            result['msg'] = "No valid input data specified"
            return result

        model = kwargs.get('model')
        if model == None:
            result['msg'] = "No valid model specified"
            return result
        else:
            success = self.set_model(task, model)
            if success:
                model = self.get_model()
            else:
                result['msg'] = "No valid model specified"
                return result

        # call OpenAI API
        try:
            if task == 'completion':
                result['text'] = self.call_completion_api(prompt, model)
                result['success'] = True
            elif task == 'embedding':
                result['embedding'] = self.call_embedding_api(prompt, model)
                result['success'] = True
            else:
                result['msg'] = "Unknown task specified, cannot proceed"
        except:
            result['msg'] = "Something went wrong when calling the API"

        return result

    def call_completion_api(self, prompt, model):
        """
        Make a call to OpenAI API to get the text completion
        """
        openai.api_key = self.api_key

        max_tokens = round(len(prompt)*1.5)
        if model == self.default_code_model:
            max_tokens = 2047

        response = openai.Completion.create(
          model=model,
          prompt=prompt,
          temperature=0,
          max_tokens=max_tokens,
          top_p=1.0,
          frequency_penalty=0.0,
          presence_penalty=0.0,
          stop=["#", ";"]
        )

        text = response.choices[0]['text']

        return text

    def call_embedding_api(self, prompt, model):
        """
        Make a call to OpenAI API to get the embedding
        """
        openai.api_key = self.api_key

        response = openai.Embedding.create(
          model=model,
          input=prompt
        )

        embedding = response['data'][0]['embedding']

        return embedding


if __name__ == "__main__":

    cred = get_credentials_by_name('openai-api')

    prompt1 = "test"
    prompt2 = "Create a SQL request to find all users who live in California and have over 1000 credits"
    prompt3 = """### Postgres SQL tables, with their properties:
#
# Employee(id, name, department_id)
# Department(id, name, address)
# Salary_Payments(id, employee_id, amount, date)
#
### A query to list the names of the departments which employed more than 10 employees in the last 3 months"""

    generator = LLMTextGenerator(cred)
    result = generator.generate_embedding(prompt=prompt3)

    if result['success'] == True:
        print (result['embedding'])
    else:
        print (result['msg'])



"""
Reusable library of modules (e.g., base classes and algorithms) 
to be incorporated into transforms

"""
from .changepoints import *
from .llmtextgen import *
from .anonymizer import *
from .timeseries_forecasting import *



"""
etl_job.py
~~~~~~~~~~

This Python module contains an example Apache Spark ETL job definition
that implements best practices for production ETL jobs. It can be
submitted to a Spark cluster (or locally) using the 'spark-submit'
command found in the '/bin' directory of all Spark distributions
(necessary for running any Spark job, locally or otherwise). For
example, this example script can be executed as follows,

    $SPARK_HOME/bin/spark-submit \
    --master spark://localhost:7077 \
    --py-files packages.zip \
    --files configs/etl_config.json \
    jobs/etl_job.py

where packages.zip contains Python modules required by ETL job (in
this example it contains a class to provide access to Spark's logger),
which need to be made available to each executor process on every node
in the cluster; etl_config.json is a text file sent to the cluster,
containing a JSON object with all of the configuration parameters
required by the ETL job; and, etl_job.py contains the Spark application
to be executed by a driver process on the Spark master node.

For more details on submitting Spark applications, please see here:
http://spark.apache.org/docs/latest/submitting-applications.html

Our chosen approach for structuring jobs is to separate the individual
'units' of ETL - the Extract, Transform and Load parts - into dedicated
functions, such that the key Transform steps can be covered by tests
and jobs or called from within another environment (e.g. a Jupyter or
Zeppelin notebook).
"""

import os
import sys
import json
import argparse
import traceback

from dependencies import enrich


def add_data(config):

    sqlsc = config["sqlsc"]
    root = config["root"]
    data = {}
    for name, path in config["data"].items():
        filename = os.path.join(root, path)
        df = sqlsc.read.csv(filename, header=True, mode="DROPMALFORMED", sep=",")

        data[name] = df.rdd

    config["rdds"] = data


def run(config):

    from dependencies import supernova

    add_data(config)

    params = config["params"]

    # Initialize the tagger...
    tagging_extra = supernova.initialize(params)

    rdd = config["rdds"]["members"]

    def reducefunc(g):
        key = g[0]
        rows = g[1]
        rows = list(rows)

        # All elements are text. So fix the types..

        # Now run the tagger..
        records = supernova.process(rows, tagging_extra)

        print("LOOKING AT KEY", key)
        print("LOOKING AT ROWS", rows)
        result = (key, len(rows))
        print("RESULT", result)
        return {"key": key, "len": len(rows)}

    def grouping(row):
        return tuple([row[x] for x in ["MEMBERSHIP_ID", "bdate", "storenmbr"]])

    grouped = rdd.groupBy(grouping)

    collected = grouped.map(reducefunc)
    print("GROUPED COLLECT", collected.take(2))

    collected = collected.collect()

    print("COLLECTED", collected)


# entry point for PySpark ETL application
if __name__ == "__main__":

    # => Load configuration and pass any additional parameters...
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", required=True, help="Configuration file")
    args = parser.parse_args()
    config = json.load(open(args.config))
    config["root"] = os.path.dirname(args.config)

    # Start the spark session
    enrich.start_spark_session(config)

    # Run the code...
    try:
        run(config)
    except:
        traceback.print_exc()

    # Shut down
    enrich.stop_spark_session(config)



import os
import sys
import json
import trp
import boto3
from datetime import datetime
import tempfile


from pdf2image import convert_from_path

class ExtractorBase():
    """
    Base class for extractors
    """
    pass

class AWSTextactor(ExtractorBase):
    """
    Class to use Textractor to process images...

    There are two implementations:
    (1) Low-level - get_text etc
    (2) TRP-library-based - doc_to_tabletext, doc_to_linetext

    https://github.com/mludvig/amazon-textract-parser
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/textract/client/analyze_document.html
    https://docs.aws.amazon.com/textract/latest/dg/examples-export-table-csv.html
    """

    def __init__(self, cred=None):
        self.cred = cred

    def get_text(self, result, blocks_map):
        """

        """
        text = ''
        if 'Relationships' in result:
            for relationship in result['Relationships']:
                if relationship['Type'] == 'CHILD':
                    for child_id in relationship['Ids']:
                        word = blocks_map[child_id]
                        if word['BlockType'] == 'WORD':
                            text += word['Text'] + ' '
                        if word['BlockType'] == 'SELECTION_ELEMENT':
                            if word['SelectionStatus'] =='SELECTED':
                                text +=  'X '
        return text

    def get_table_rows(self, table_block, blocks_map):
        """
        Get text from the cells
        """

        rows = {}
        for relationship in table_block['Relationships']:
            if relationship['Type'] == 'CHILD':
                for child_id in relationship['Ids']:
                    cell = blocks_map[child_id]
                    if cell['BlockType'] == 'CELL':
                        row_index = cell['RowIndex']
                        col_index = cell['ColumnIndex']

                        if row_index not in rows:
                            # create new row
                            rows[row_index] = {}

                        # get the text value
                        rows[row_index][col_index] = self.get_text(cell, blocks_map)
        return rows

    def get_one_table(self, table_block, blocks_map):
        """
        Get one table's content by looking at the children
        """
        rows = self.get_table_rows(table_block, blocks_map)

        result = []
        for row_index, cols in rows.items():
            row = []
            for col_index, text in cols.items():
                row.append(text)
            result.append(row)

        return result

    def get_all_tables(self, blocks):

        tables = {}

        # => Which blocks have tables?
        blocks_map = {}
        table_blocks = []
        for block in blocks:
            blocks_map[block['Id']] = block
            if block['BlockType'] == "TABLE":
                table_blocks.append(block)

        if len(table_blocks) <= 0:
            return tables

        for index, table_block in enumerate(table_blocks):
            tableid = f"table_{index+1}"
            tables[tableid] = self.get_one_table(table_block, blocks_map)

        return tables


    def get_all_signatures(self, blocks):

        signatures = {}

        # => Which blocks have tables?
        blocks_map = {}
        signature_blocks = []
        for block in blocks:
            blocks_map[block['Id']] = block
            if block['BlockType'] == "SIGNATURE":
                signature_blocks.append(block)

        if len(signature_blocks) <= 0:
            print("NO signatures found")
            return signatures

        for index, sig_block in enumerate(signature_blocks):
            sigid = f"sig_{index+1}"
            print(json.dumps(sig_block, indent=4))
            signatures[sigid] = {
                'confidence': sig_block['Confidence']
            }
            if 'Text' in sig_block:
                signatures[sigid].update({
                    'text': sig_block['Text'],
                    'type': sig_block['TextType'],
                })

        return signatures

    def get_blocks(self, content):

        # Get the text blocks
        blocks = content['Blocks']

        return blocks

    def get_all_content(self, content):

        featuretypes=['TABLES', 'SIGNATURES']

        if isinstance(content, str):
            img_test = open(content, 'rb').read()
            bytes_test = bytearray(img_test)
        elif isinstance(content, bytes):
            bytes_test = bytearray(content)
        else:
            raise Exception("Content should be filename or byte array")

        # process using image bytes
        # get the results
        session = boto3.Session()
        client = session.client('textract')
        response = client.analyze_document(Document={'Bytes': bytes_test},
                                           FeatureTypes=featuretypes)

        return response

    def doc_to_signatures(self, textract_data):
        """
        Take a textract response and extract all signatures
        """
        content = textract_data['content']
        filename = textract_data['filename']
        pageno = textract_data['page']

        # check for signatures
        signatures = []
        signo = 0
        for item in content["Blocks"]:
            if item["BlockType"] == "SIGNATURE":
                one_sig = {
                    "id": f"{filename}-p{pageno}-s{signo}",
                    "page": pageno,
                    "confidence": item["Confidence"],
                    "location": item["Geometry"]
                }
                signatures.append(one_sig)
                signo += 1

        return signatures

    def doc_to_tabletext(self, textract_data):
        """
        Take a textract response and extract all tables
        """
        content = textract_data['content']
        filename = textract_data['filename']
        pageno = textract_data['page']

        # convert to Document format
        doc = trp.Document(content)

        # extract all tables
        tables = []
        tableno = 0
        for page in doc.pages:
            # for each table
            for table in page.tables:
                table_rows = []
                # for each row
                for row in table.rows:
                    table_rows.append("\t".join([cell.text for cell in row.cells]))

                # we now have all the text for one table
                table_text = "\n".join(table_rows)

                # construct one table
                one_table = {
                    "id": f"{filename}-p{pageno}-t{tableno}",
                    "text": table_text
                }
                tables.append(one_table)

                tableno += 1

        return tables

    def doc_to_linetext(self, textract_data):
        """
        Take a textract response and extract all tables
        """
        content = textract_data['content']
        filename = textract_data['filename']
        pageno = textract_data['page']

        blocks = content['Blocks']
        text = ""
        for item in blocks:
            if item["BlockType"] == "LINE":
                text += item["Text"] + "\n"
        text = {
            "id": f"{filename}-p{pageno}-text",
            "text": text.strip()
        }

        return text

    def extract_and_organize(self, content):

        # First get the structure of the document
        response = self.get_all_content(content)

        # Incorporate the blocks..
        blocks = self.get_blocks(response)
        tables = self.get_all_tables(blocks)
        signatures = self.get_all_signatures(blocks)

        return {
            "timestamp": datetime.now().isoformat(),
            'metadata': {
                'blocks': blocks
            },
            'tables': tables,
            'signatures': signatures
        }

    def extract(self, content):
        # Get the full structured content of the document
        response = self.get_all_content(content)
        return response


def extract_text_from_file(filename, provider="aws"):

    if provider == "aws":
        handler = AWSTextactor()
    else:
        raise Exception("Only aws is supported as a provider for now")

    allresults = []
    if filename.lower().endswith("pdf"):
        with tempfile.TemporaryDirectory() as path:
            images_from_path = convert_from_path(filename,
                                                 output_folder=path,
                                                 fmt='png',
                                                 poppler_path="")
            print (f"Extracting text from: {filename}")
            for idx, img in enumerate(sorted(os.listdir(path))):
                print (f"..page {idx}")
                response = handler.extract(os.path.join(path, img))
                result = {
                    "content": response,
                    "filename": filename,
                    "page": idx,
                }
                result['signatures'] = handler.doc_to_signatures(result)
                result['tables'] = handler.doc_to_tabletext(result)
                result['text'] = handler.doc_to_linetext(result)
                allresults.append(result)

    else:
        result = handler.extract(filename)
        result['page'] = 1
        allresults.append(result)

    return allresults



import os
import sys
import json
import copy
import tempfile
import shutil
import time
import glob
import re
import traceback
import subprocess
import numpy as np
import pandas as pd
from enrichsdk import S3Mixin, Compute
from enrichsdk.contrib.lib.assets import AnonymizerMixin

from datetime import datetime, date, timedelta
from dateutil import parser as dateparser, relativedelta
import logging

logger = logging.getLogger("app")

from enrichsdk.utils import (get_lineage_of_query,
                             note,
                             SafeEncoder,
                             get_yesterday,
                             get_today)

class InMemoryQueryExecutorBase(AnonymizerMixin, Compute):
    """
    Base class for an InMemory QueryExecutor transform. This is useful
    to run queries against backends such as backends such as
    mysql

    Features of transform baseclass include:

        * Support multiple query engines (via SQLAlchemy)
        * Support templatized execution
        * Support arbitrary number of queries
        * Supports a generator function to generate per-interval queries

    Configuration looks like::

        ...
        "args": {
            "cleanup": False,
            "force": True,
            "targets": "all",
            "start_date": "2020-08-01",
            "end_date": "2020-08-03",
        }

     Specs

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "QueryExecutorBase"
        self.description = "Execute queries against backends"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def preload_clean_args(self, args):
        """
        Check validity of the args
        """
        args = super().preload_clean_args(args)

        if ("start_date" not in args) or ("end_date" not in args):
            raise Exception("Start or end of timeframe missing")

        try:
            start = dateparser.parse(args["start_date"]).date()
            args["start_date"] = start
            end = dateparser.parse(args["end_date"]).date()
            args["end_date"] = end
        except:
            logger.exception(
                "Invalid start_date or end_date", extra={"transform": self.name}
            )
            raise Exception("Invalid start/end datetime specified")

        if (
            ("targets" not in args)
            or (not isinstance(args["targets"], str))
            or (len(args["targets"]) == 0)
        ):
            raise Exception("Invalid list of query names specified")

        # Include force
        force = str(args["force"]).lower().strip()
        force = force == "true"
        args["force"] = force

        # Clean the list of names...
        targets = args["targets"].split(",")
        targets = [n.strip() for n in targets if len(n.strip()) > 0]
        args["targets"] = [n for n in targets if len(n) > 0]

        return args

    def get_supported_extra_args(self):
        """
        Look at the specs to generate a list of options that
        can be presented to the end-ser
        """

        # Collect specs first..
        specs = self.get_sql_specs()

        # Compute the targets
        targets = ["all"]  # default
        for s in specs:
            if not s.get("enable", True):
                continue
            categories = s.get("categories", s.get('category', []))
            if isinstance(categories, str):
                categories = [categories]
            for c in categories:
                if c not in targets:
                    targets.append(c)
        for s in specs:
            name = s["name"]
            if name not in targets:
                targets.append(name)
        targets = "|".join(targets)

        # Now construct the args dynamically
        remaining = self.supported_extra_args
        return [
            {
                "name": "targets",
                "description": f"What all to run. Specify multiple with comma separating names ({targets})",
                "default": "all",
                "required": False,
            },
            {
                "name": "force",
                "description": "Force execution",
                "default": "False",
                "required": False,
            },
            {
                "name": "start_date",
                "description": "Start of the time window",
                "default": get_yesterday(),
                "required": True,
            },
            {
                "name": "end_date",
                "description": "End of the time window",
                "default": get_today(),
                "required": True,
            },
        ] + remaining

    def update_frame(self, name, engine, sql, df, dependencies=[]):
        """
        Note the lineage for each output file.
        """

        # Check if it has already been registered
        if self.state.has_frame(name):
            return

        # Get the default database
        database = engine.url.database

        # Insert extra dependencies
        try:
            dependencies += get_lineage_of_query(engine, sql)
        except:
            dependencies = []
            logger.warning("Unable to get lineage",
                             extra={
                                 'transform': self.name,
                                 'data': f"SQL being checked:\n {sql}"
                             })

        # Generate column information...
        columns = self.get_column_metadata(name, df)

        ## => Gather the update parameters
        updated_detail = {
            "df": df,
            "description": f"Output for query {name}",
            "transform": self.name,
            "frametype": "pandas",
            "params": [
                {
                    "type": "compute",
                    "columns": columns,
                },
            ],
        }

        if len(dependencies) > 0:
            lineage = {"type": "lineage", "dependencies": dependencies}
            updated_detail['params'].append(lineage)

        # Dump it into the shared state
        self.state.update_frame(name, updated_detail, create=True)

    def generate_paramsets(self, spec, start_date, end_date):

        duration = spec.get("paramsets_duration", "full")
        window = spec.get("paramsets_window", 0)

        if isinstance(start_date, str):
            start_date = parser.parse(start_date).date()

        if isinstance(end_date, str):
            end_date = parser.parse(end_date).date()

        if duration not in ["day", "full"]:
            raise Exception("Only day and full are supported. full is default")

        if duration == "full":
            return [
                {
                    "start_date": start_date.isoformat(),
                    "end_date": end_date.isoformat(),
                    "start_date_minus7": (
                        start_date + relativedelta.relativedelta(days=-7)
                    ).isoformat(),
                    "start_date_minus1": (
                        start_date + relativedelta.relativedelta(days=-1)
                    ).isoformat(),
                    "start_date_minus6_months": (
                        start_date + relativedelta.relativedelta(months=-6)
                    ).isoformat(),
                    "end_date_plus1": (
                        end_date + relativedelta.relativedelta(days=1)
                    ).isoformat(),
                    "end_date_minus1": (
                        end_date + relativedelta.relativedelta(days=-1)
                    ).isoformat(),
                    "end_date_minus7": (
                        end_date + relativedelta.relativedelta(days=-7)
                    ).isoformat(),
                    "end_date_minus6_months": (
                        end_date + relativedelta.relativedelta(months=-6)
                    ).isoformat(),
                }
            ]

        # There may be more. So leaving it here...
        if duration == "day":

            # Parse the dates
            if start_date > end_date:
                start_date, end_date = end_date, start_date

            # Sanity check
            diff = end_date - start_date
            if diff.days > 100:
                raise Exception("Date range should be < 100 days")

            # Generate one for each day
            paramsets = []
            curr_date = start_date
            while curr_date <= end_date:
                paramsets.append(
                    {
                        "start_date": curr_date.isoformat(),
                        "end_date": (
                            curr_date + relativedelta.relativedelta(days=window)
                        ).isoformat(),
                        "start_date_minus7": (
                            curr_date + relativedelta.relativedelta(days=-7)
                        ).isoformat(),
                        "start_date_minus1": (
                            curr_date + relativedelta.relativedelta(days=-1)
                        ).isoformat(),
                        "start_date_plus1": (
                            curr_date + relativedelta.relativedelta(days=1)
                        ).isoformat(),
                        "end_date_minus1": (
                            curr_date + relativedelta.relativedelta(days=-1)
                        ).isoformat(),
                        "end_date_plus1": (
                            curr_date + relativedelta.relativedelta(days=1)
                        ).isoformat(),
                        "end_date_minus7": (
                            curr_date + relativedelta.relativedelta(days=-7)
                        ).isoformat(),
                        "end_date_minus6_months": (
                            curr_date + relativedelta.relativedelta(months=-6)
                        ).isoformat(),
                    }
                )
                curr_date += relativedelta.relativedelta(days=1)

            return paramsets

        # Default...
        raise Exception(f"Unknown duration: {duration}")

    def get_sql_specs(self):
        """
        Return a list of query specifications.

        Specification: A list of dictionaries. Each dict has

          * name: Name of the specification
          * sql: SQL template
          * categories: String or a list of strings indicating specification groups
          * segment: How to split the dataframe resulting from query execution. Could be none ('complete' as the default name), string (column name) or a callback that generates a { name: df } map
          * paramsets_duration: each instance for one 'day' or a window of days (defined below)
          * paramsets_window: each instance translates into date range for each instance of parameters.

        Examples::

           Simple:
             {
                 "name": "txn_value",
                 "sql": "txn_value.sql",
                 "segment": "global_date",
             }

           Simple:

             {
                 "categories": ["kyc"],
                 "name": "kyc_txn_summary",
                 "sql": "kyc_txn_summary.sql",
                 "segment": complex_split_callbak,
                 "paramsets_duration": "day",
                 "retries": 3,
             },

        """

        self.get_specs()

    def get_specs(self):
        """
        Use get_sql_specs instead.

        .. warning::
            .. deprecated:: 2.6.0

        """

        return []

    def get_registry(self):
        """
        Build a registry and return
        """
        return None

    def get_engine(self, spec):
        """
        Build and return an engine for a given specification.
        """
        raise Exception("Construct sqlalchemy engine")

    def generic_clean(self, df):
        """
        Do a high level clean of the query result before
        doing a query-specific clean
        """
        return df

    def get_specs_from_sqls(self, sqldir):
        """
        Helper function. Load specifications from the SQLs.
        """
        specs = []

        files = glob.glob(sqldir + "/*.sql")
        for f in files:
            name = os.path.basename(f).replace(".sql", "")
            sql = open(f).read()

            # Specify the split in the SQL itself..
            segment = None
            match = re.search(r"-- segment:: (\S+)", sql)
            if match is not None:
                segment = match.group(1).strip()

            match = re.search(r"-- name:: (\S+)", sql)
            if match is not None:
                name = match.group(1).strip()

            match = re.search(r"-- engine:: (\S+)", sql)
            if match is not None:
                engine = match.group(1).strip()

            specs.append(
                {"name": name, "sql": sql, "segment": segment, "engine": engine}
            )

        return specs

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the registry
        self.registry = self.get_registry()

        # => Initialize anonymization if required
        if 'anonymization' in self.args:
            self.anonymize_init(self.args['anonymization'])

        # List of specification names
        targets = self.args["targets"]

        # Get specs..
        specs = self.get_sql_specs()

        logger.debug(f"Specs found: {len(specs)}", extra={"transform": self.name})
        # Now iterate through the specs.
        for spec in specs:
            try:

                name = spec["name"]
                categories = spec.get("categories",
                                      spec.get('category', ['all']))
                if isinstance(categories, str):
                    categories = [categories]
                table = spec.get("table", name)
                cond = spec.get("cond", "")
                retries = spec.get("retries", 1)

                # To take care of the logging in case of exception
                msg = f"Name: {name}\n"

                # Check if this has been requested?
                if all([c not in targets for c in categories]) and (
                    name not in targets
                ):
                    continue

                logger.debug(
                    f"Executing {spec['name']}",
                    extra={
                        "transform": self.name,
                        "data": json.dumps(spec, indent=4, cls=SafeEncoder),
                    },
                )

                sql_template = spec["sql"]

                files = []

                paramsets = self.generate_paramsets(
                    spec, self.args["start_date"], self.args["end_date"]
                )

                for params in paramsets:

                    status = []

                    msg = f"Params: {params}\n"
                    msg += f"Insert Table: {table} with {cond}\n"

                    # Now log the SQL
                    sql = sql_template % params
                    msg += "SQL:\n{}\n".format(sql)

                    # Get the engine for a given spec
                    engine = self.get_engine(spec)

                    segmentcol = spec.get("segment", None)

                    tryno = 1
                    while True:
                        if tryno > retries:
                            raise Exception("Exceeded max retries")

                        try:
                            df = pd.read_sql(sql, engine)
                            break
                        except:
                            logger.exception(
                                f"Failed Query: {name} (try {tryno})",
                                extra={"transform": self.name, "data": msg},
                            )
                        tryno += 1
                        time.sleep(30)

                    # Do some basic cleaning. int becomes float
                    df = self.generic_clean(df)

                    msg += f"Segment: {segmentcol} (Initial split)\n"
                    msg += "Records: {}\n".format(df.shape[0])
                    msg += "Columns: {}\n".format(", ".join(df.columns))
                    msg += "Dtypes: " + df.dtypes.to_string() + "\n"

                    skip_empty = spec.get("skip_empty", True)
                    if len(df) == 0:
                        # no data returned...
                        if skip_empty:
                            logger.warning(
                                f"Completed {name} {params['start_date']} No data",
                                extra={"transform": self.name, "data": msg},
                            )
                            continue
                        else:
                            logger.warning(
                                f"{name} {params['start_date']} No data",
                                extra={"transform": self.name, "data": msg},
                            )

                    if ((len(df) == 0) and (not callable(segmentcol))):
                        msg = """Dont know how to handle an empty dataframe. Not sure what columns should be included with what values. segmentcol should be a callable""" + msg
                        logger.warning(f"{name} {params['start_date']} Skipping",
                                       extra={
                                           "transform": self.name,
                                           "data": msg
                                       })
                        continue

                    # First gather a map of segments
                    filemap = {}
                    if segmentcol is None:
                        # Whole thing is one segment
                        filemap["complete"] = df
                    elif isinstance(segmentcol, str):
                        # Split by column name...
                        segments = list(df[segmentcol].unique())
                        msg += f"Segments: {len(segments)} ({segmentcol})\n"
                        for segment in segments:
                            try:
                                df1 = df[df[segmentcol] == segment]
                                segment = str(segment)
                                filemap[segment] = df1
                            except:
                                pass
                    elif callable(segmentcol):
                        # Custom split of the dataframe...
                        filemap = segmentcol(self, spec, params, df)
                        msg += f"Segments: {len(filemap)}\n"
                    else:
                        raise Exception(f"Unhandled segment definition: {segmentcol}")

                    # => Process each segment obtained from
                    # the previous step...
                    for segment, df1 in sorted(filemap.items()):

                        # Add note about what is being stored..
                        msg += f"[{segment}] {df1.shape[0]} records\n"

                        # Clean the output data...
                        try:
                            if "clean" in spec:
                                callback = spec["clean"]["callback"]
                                if callable(callback):
                                    clean_msg, df1, clean_files = callback(
                                        self, segmentcol, segment, df1, spec
                                    )
                                    if len(clean_msg) > 0:
                                        msg += f"[{segment}] " + clean_msg + "\n"
                                    files += clean_files
                        except Exception as e:
                            #traceback.print_exc()
                            msg += str(e)
                            raise

                        # Store in database...
                        try:
                            extra_dependencies = []
                            if "store" in spec:
                                # Separate storage handler..
                                callback = spec["store"]["callback"]
                                if callable(callback):
                                    store_msg, store_dependencies, store_files = callback(
                                        self, segmentcol, segment, df1, spec
                                    )
                                    if len(store_msg) > 0:
                                        msg += f"[{segment}] " + store_msg + "\n"
                                    extra_dependencies += store_dependencies
                                    files += store_files

                        except Exception as e:
                            #traceback.print_exc()
                            msg += str(e)
                            raise

                        # Handle a default store for all specs, segments
                        try:


                            # => Anonymize the data
                            if hasattr(self, 'anonargs'):
                                anon_df1 = self.anonymize_target(spec['name'], df=df1)
                            else:
                                anon_df1 = None

                            # Store in s3 etc.
                            store_msg, store_dependencies, store_files  = self.store(
                                segmentcol, segment, df1, anon_df1, spec
                            )
                            if len(store_msg) > 0:
                                msg += f"[{segment}] " + store_msg

                            extra_dependencies += store_dependencies
                            files += store_files

                            # update lineage
                            self.update_frame(
                                name, engine, sql, df1, extra_dependencies
                            )


                        except Exception as e:
                            #traceback.print_exc()
                            msg += "[{}] Exception {}\n".format(segment, traceback.format_exc()) + "\n"

                    logger.debug(
                        f"Completed {name} {params['start_date']}",
                        extra={"transform": self.name, "data": msg},
                    )

                # Make note of it.
                dataset = self.registry.find(spec['name'])
                if ((dataset is not None) and (len(files) > 0)):
                    metadata = { 'files': files}
                    self.registry.access(dataset, metadata, nature='write')

            except:
                #traceback.print_exc()
                # Exception for each spec.
                logger.exception(
                    f"Unable to run query: {name}",
                    extra={"transform": self.name, "data": msg},
                )
                msg = ""
                continue

        self.add_marker(state)


        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



"""
Store
-----

Interface to the key-value store backends such as Redis and Elastic
Search.

"""
import json
import redis
import traceback
import logging

logger = logging.getLogger("app")


class BackendBase(object):
    """
    Base class to handle KV store interface
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the db object

        Args:
            config (dict): Configuration
        """
        self.config = kwargs.pop("config", {})

    def connect(self):
        """
        Initialize the connection to the backend
        """
        pass

    def post(self, key, value):
        """
        Post a kv combination to the backend

        Args:
            key (str): key
            value (dict): document to the posted
        """
        raise Exception("Not implemented")

    def get(self, key):
        """
        Lookup a key at the backend

        Args:
            key (str): key
        """
        raise Exception("Not implemented")


class RedisBackend(BackendBase):
    """
    Redis Backend

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the db object. Pass redis
        connection parameters as 'params'
        attribute in the config.

        Args:
            config (dict): Configuration
        """
        super().__init__(*args, **kwargs)
        self.r = None

    def connect(self):
        """
        Connect to Redis backend
        """
        params = self.config.get("params", {})
        self.r = redis.Redis(**params)

    def get(self, key):
        """
        Lookup a key at the backend

        Args:
            key (str): key
        """
        try:
            result = self.r.get(key)
            if result is None:
                return {}
            return json.loads(result.decode("utf-8"))
        except:
            traceback.print_exc()
            return {}

    def post(self, key, value):
        """
        Post a kv combination to the backend. If value
        is not a string, then it is serialized using
        json

        Args:
            key (str): key
            value (dict): document to the posted
        """
        try:
            if not isinstance(value, str):
                result = json.dumps(value)
            self.r.set(key, result)
        except:
            logger.exception("Unable to post to redis server")
            raise



import os
import sys
import json
import copy
import inspect
import glob
import pytz
import getpass
import platform
import traceback
import distro
from datetime import datetime
from dateutil import parser as date_parser

import pandas as pd

from ..core.state import EnrichStateBase
from ..core.frames import PandasDataFrame, DictDataFrame
from ..lib import read_siteconf, read_versionmap
from ..lib.context import Context
import logging

from enrichsdk.utils import SafeEncoder

logger = logging.getLogger("app")


class MockRun(object):
    def __init__(self, *args, **kwargs):
        self.state = kwargs.pop("state")
        self.files = []

    def load_files(self):
        frames = self.state.get_frame_list()
        for f in frames:
            detail = self.state.get_frame(f)
            detail["name"] = f
            self.files.append(detail)

    def list_files(self):
        return [f["name"] for f in self.files]

    def get_file_hash(self):
        return {f["name"]: f for f in self.files}


class MockRunManager(object):
    def __init__(self, *args, **kwargs):
        self.runs = []

    def load_runs(self, runs, state):
        for r in runs:
            status = r["status"]
            runid = r["runid"]
            run = MockRun(state=state)
            run.status = r["status"]
            run.runid = r["runid"]
            run.start_time = date_parser.parse(r["start_time"])
            run.end_time = date_parser.parse(r["end_time"])

            self.runs.append(run)

    def get_last_run(self, status="success"):
        for r in self.runs:
            if r.status == status:
                return r
        return run

    def load(self):
        pass


class MockState(EnrichStateBase):
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.name = "ExperimentalPipeline"
        self.runid = "exp-pipeline-20180922-192321"
        self.state = {}
        self.notes = []
        user = getpass.getuser()
        self.stats = {
            "platform": {
                "node": platform.node(),
                "os": platform.system(),
                "release": platform.release(),
                "processor": platform.processor(),
                "python": platform.python_version(),
                "distribution": (distro.id().title(), distro.version(), distro.name())
            },
            "user": user,
        }

    def make_note(self, text):
        self.notes.append(text)

    def make_dataset_access_note(self, dataset, metadata, nature='read-write'):
        if dataset is None:
            return

        self.dataset_accesses.append({
            'timestamp': datetime.now(),
            'name': dataset.name,
            'nature': nature,
            'description': dataset.description,
            'metadata': metadata
        })

    def make_performance_note(
        self, name, description, condition, priority=1, message=""
    ):
        self.performance_summary.append(
            {
                "name": name,
                "description": description,
                "condition": condition,
                "priority": priority,
                "message": message,
            }
        )

    def update_frame(self, framename, details, create=False):

        details["name"] = framename

        for d in ["df", "transform"]:
            if d not in details:
                raise Exception("Missing frame detail: {}".format(d))
        if "frametype" not in details:
            details["frametype"] = "pandas"

        if "description" not in details:
            details["description"] = ""

        if "category" not in details:
            details["category"] = "default"

        if "stages" not in details:
            details["stages"] = []
        if isinstance(details["transform"], str):
            details["stages"].append(details["transform"])
        else:
            details["stages"].extend(details["transform"])

        if "params" not in details:
            details["params"] = []
        elif isinstance(details["params"], dict):
            details["params"] = [details["params"]]
        elif isinstance(details["params"], list):
            for p in details["params"]:
                if not isinstance(p, dict):
                    raise Exception("Invalid params. List items should be dicts")
        else:
            raise Exception("Invalid params. Expected list or dict")

        if "history" not in details:
            details["history"] = []
        elif isinstance(details["history"], dict):
            details["history"] = [details["history"]]

        if framename not in self.state:
            frame = self.state[framename] = details
        else:
            frame = self.state[framename]
            frame["df"] = details["df"]
            frame["source"] = details["transform"]
            frame["stages"].append(details["transform"])
            frame["history"].extend(details["history"])

            # Handle description
            if (isinstance(details["description"], str)) and (
                len(details["description"]) > 0
            ):
                if frame["description"] != "":
                    frame["description"] += ". "
                frame["description"] += details["description"]

            # Incorporate params
            if "params" in details:
                if isinstance(details["params"], dict):
                    frame["params"].insert(0, details["params"])
                elif isinstance(details["params"], list):
                    for p in details["params"]:
                        if not isinstance(p, dict):
                            raise Exception("Invalid params type")
                        if len(p) == 0:
                            continue
                        frame["params"].append(p)
                else:
                    raise Exception("Unknown params")

            if (frame["category"] == "default") and (details["category"] != "default"):
                frame["category"] = details["category"]

        frame['last_update'] = datetime.now().isoformat()

    def has_frame(self, framename):
        return framename in self.state

    def get_frame(self, framename):
        if framename not in self.state:
            raise Exception("Cannot find the required frame")
        return self.state[framename]

    def get_frame_list(self, category=None):
        allframes = list(self.state.keys())
        if category is None:
            return allframes

        frames = []
        for f in allframes:
            if self.state[f]["category"] == category:
                frames.append(f)

        return frames

    def reached_stage(self, frame, name):
        return (frame in self.state) and (name in self.state[frame]["stages"])

    def add_expectations(self, transform, frame, expectations):
        if not isinstance(expectations, list):
            raise Exception("Expectations should be a list")

        for e in expectations:
            for col in ["expectation", "description", "passed"]:
                if col not in e:
                    raise Exception(
                        "Expectation results are missing col {}".format(col)
                    )

        # => Insert metadata and store
        expanded = []
        for e in expectations:
            e = copy.copy(e)
            e.update(
                {
                    "version": "v1",
                    "timestamp": self.config.now().isoformat(),
                    "frame": frame,
                    "meta": {
                        "node": self.stats["platform"]["node"],
                        "pipeline": self.name,
                        "transform": transform.name,
                        "runid": self.runid,
                        "start_time": self.start_time.isoformat(),
                    },
                }
            )
            self.expectations.append(e)


class MockConfig(object):
    def __init__(self, context={}, *args, **kwargs):

        # Initialize the context
        if len(context) == 0:
            context = Context()
            context.set_env()
            context = context.asdict()

        self.name = "TestConf"
        self.display = self.name
        self.runid = "run-2015021"
        self.description = "Experimental pipeline for testing transforms"
        self.context = context
        self.runmanager = MockRunManager()
        self.dataframes = {"pandas": PandasDataFrame(), "dict": DictDataFrame()}
        self.tz = pytz.timezone("Asia/Kolkata")
        self.ts = datetime.now(self.tz).replace(microsecond=0).isoformat()
        self.enable_extra_args = True
        self.readonly = False
        self.data_root = self.context["ENRICH_TEST"]
        self.customer_root = self.context["ENRICH_CUSTOMERS"]
        self.enrich_customers = self.context["ENRICH_CUSTOMERS"]
        self.enrich_root = self.context["ENRICH_ROOT"]
        self.enrich_run_root = self.context["ENRICH_ROOT"]
        self.enrich_data_dir = self.data_root
        self.accessed_files = []
        self.supported_extra_args = []
        self.usecase = {"org": {"name": "Unknown", "customer": "unknown"}}
        # Add defaults
        for attr in [
            "data_root",
            "enrich_data_dir",
            "inputdir",
            "outputdir",
            "statedir",
        ]:
            setattr(self, attr, self.data_root)

        self.state = MockState(config=self)

    def get_usecase(self):
        return self.usecase.get("org", {}).get("name", "Unknown")

    def get_customer(self):
        return self.get_usecase()

    def load_usecase(self, path):

        enrichjson = None
        enrich_customers = self.enrich_customers

        if "ENRICHJSON" in os.environ:
            enrichjson = os.environ["ENRICHJSON"]
            enrichjson = os.path.abspath(enrichjson)
        elif "ENRICHJSON" in self.context:
            enrichjson = os.path.abspath(self.context["ENRICHJSON"])
        elif os.path.exists(enrich_customers):
            # Try inferring from the location.
            enrichjson = None
            path = os.path.realpath(path)
            for c in os.listdir(enrich_customers):
                c = os.path.realpath(os.path.join(enrich_customers, c))
                if path.startswith(c):
                    relpath = os.path.relpath(path, start=c)
                    enrichjson = os.path.join(
                        c, relpath.split(os.sep)[0], "enrich.json"
                    )
                    break

        if (enrichjson is None) or (not os.path.exists(enrichjson)):
            logger.warning("Unable to find enrich.json")
            return {}

        try:
            enrich = json.load(open(enrichjson))
        except:
            raise Exception("enrich.json should be valid json file")

        self.usecase = enrich

    def get_cmdline_args(self):
        return {}

    def configure(self, params):
        for var in ["enable_extra_args"]:
            if var in params:
                setattr(self, var, params[var])

    def now(self):
        return datetime.now(self.tz).replace(microsecond=0)

    def get_dataframe(self, name):

        if name == "pandas":
            return self.dataframes["pandas"]
        elif name == "dict":
            return self.dataframes["dict"]
        else:
            raise Exception("Unsupported dataframe type")

    def get_extra(self, extra={}):
        return extra

    def get_file(self, template, abspath=True, create_dir=False, extra={}):

        if template is None:
            return None

        now = self.now()
        dt = now.strftime("%Y-%m-%d")
        dtime = now.strftime("%Y%m%d-%H%M%S")
        node = platform.node()

        try:
            params = {
                "name": self.name,
                "outputdir": self.outputdir,
                "output": self.outputdir,
                "inputdir": self.inputdir,
                "input": self.inputdir,
                "statedir": self.statedir,
                "state": self.statedir,
                "runid": self.runid,
                "ts": self.ts,
                "date": dt,
                "datetime": dtime,
                "node": node,
                "data_root": self.data_root,
                "enrich_root": self.enrich_root,
                "enrich_root_dir": self.enrich_root,
                "enrich_data_dir": self.enrich_data_dir,
            }
            params.update(self.context)
            params.update(extra)
            filename = template % params
        except:
            error = "Invalid file specification: {}".format(template)
            raise Exception(error)

        if abspath:
            filename = os.path.abspath(filename)

        if create_dir:
            try:
                os.makedirs(os.path.dirname(filename))
            except:
                pass

        transform = extra.get("transform", None)
        meta = extra.get("meta", {})
        self.note_access(filename, transform, meta)

        return filename

    def note_access(self, filename, transform, meta={}):

        if transform is None:
            return

        code = None
        stack = inspect.stack()
        for s in stack:
            if (
                (".virtualenv" not in s.filename)
                and ("enrichcli" not in s.filename)
                and ("enrichsdk" not in s.filename)
            ):
                code = s.filename
                code = self.get_relative_path(code, what="enrich_root")
                code += ":" + str(s.lineno)
                break

        if os.path.abspath(filename):
            relpath = self.get_relative_path(filename, what="enrich_root")
        else:
            relpath = filename

        entry = {
            "transform": transform.name
            if not isinstance(transform, str)
            else transform,
            "filename": relpath,
            "timestamp": datetime.now().isoformat(),
            "source": code,
            "meta": meta,
        }

        self.accessed_files.append(entry)

    def get_relative_path(self, path, what="enrich_data_dir"):

        paths = {
            "enrich_root": self.enrich_root,
            "enrich_data_dir": self.enrich_data_dir,
            "outputdir": self.outputdir,
            "output": self.outputdir,
            "data_root": self.data_root,
        }
        whatroot = paths[what]
        return os.path.relpath(path, whatroot)

    def load_test_state(self, testdata):

        app = {
            "pipeline": self.name,
            "description": self.description,
            "data_root": self.data_root,
            "customer_root": self.customer_root,
            "runid": self.runid,
        }

        # Load siteconf
        if "siteconf" in testdata:
            siteconf = read_siteconf(
                testdata["siteconf"], app=app, context=self.context
            )
        elif "siteconf" in self.context:
            siteconf = read_siteconf(
                self.context["siteconf"], app=app, context=self.context
            )
        else:
            siteconf = read_siteconf(app=app, context=self.context)

        if siteconf is None:
            raise Exception("Could not read siteconf")

        self.siteconf = siteconf

        # Load versionmap
        filename = testdata.get(
            "versionmap", self.get_file("%(ENRICH_ETC)s/versionmap.json")
        )
        timestamp, versionmap = read_versionmap(filename)
        self.versionmap = versionmap

        # => Load the data frames
        expand = os.path.expandvars
        defaultdir = self.get_file("%(ENRICH_TEST)s")

        self.data_root = expand(testdata.get("data_root", defaultdir))
        self.enrich_data_dir = self.data_root
        self.inputdir = expand(testdata.get("inputdir", self.data_root))
        self.outputdir = expand(testdata.get("outputdir", self.data_root))
        self.statedir = expand(testdata.get("statedir", self.data_root))

        # Insert global supported_extra_args
        if "global" in testdata:
            if "args" in testdata["global"]:
                args = testdata["global"]["args"]
                if not isinstance(args, dict):
                    raise Exception("Global args should be a dict")
                self.supported_extra_args = [
                    {
                        "name": k,
                        "description": "Global arg {}".format(k),
                        "required": False,
                        "default": v,
                    }
                    for k, v in args.items()
                ]

        testdata = testdata.get("data", {})
        for framename, details in testdata.items():
            stages = details.get("stages", [])
            frametype = details.get("frametype", "pandas")

            # We could have potentially specified multiple files that need to be loaded
            paths = []
            transform = details["transform"]
            sources = []
            if "source" in details:
                source = details["source"]
            else:
                source = details["filename"]
            if isinstance(source, str):
                if "inputdir" not in source:
                    source = "%(inputdir)s/%(transform)s/" + source
                path = self.get_file(
                    source, extra={"inputdir": self.inputdir, "transform": transform}
                )
                sources.append(path)
                paths.extend(glob.glob(path))
            elif isinstance(source, list):
                for f in source:
                    if "inputdir" not in f:
                        f = "%(inputdir)s/%(transform)s/" + f
                    path = self.get_file(
                        f, extra={"inputdir": self.inputdir, "transform": transform}
                    )
                    sources.append(path)
                    paths.extend(glob.glob(path))
            elif isinstance(source, dict):
                for group in source:
                    for f in source[group]:
                        path = self.get_file(
                            f, extra={"transform": transform, "group": group}
                        )
                        sources.append(path)
                        paths.extend(glob.glob(path))

            if len(paths) == 0:
                logger.warning(
                    "No files to load for {}".format(framename),
                    extra={"data": "Searched " + ",".join(sources)},
                )
                continue

            _dfs = []
            for path in paths:
                if not os.path.exists(path):
                    raise Exception("Missing input file: {}".format(path))

                if frametype in ["pandas"]:
                    params = details.get("params", {})
                    df = pd.read_csv(path, **params)
                elif frametype in ["dict"]:
                    params = {}
                    df = json.load(open(path, "r"))
                _dfs.append(df)

            # => Load them ..
            if len(_dfs) == 0:
                pass
            elif len(_dfs) == 1:
                df = _dfs[0]
            else:
                if "mergedf" not in details:
                    raise Exception("Please use 'mergedf' attribute to combine")

                df = details["mergedf"](_dfs)

            details = {
                "df": df,
                "frametype": frametype,
                "stages": stages,
                "transform": details["transform"],
                "history": [{"log": "Test load of data"}],
                "params": [
                    {
                        "type": "lineage",
                        "dependencies": [
                            {"type": "file", "usage": "input", "name": paths}
                        ],
                    }
                    # {
                    #    'type': 'export',
                    #    'params': params
                    # }
                ],
            }
            self.state.update_frame(framename, details)

        # => Load the runs
        runs = testdata.get("runs", [])
        self.runmanager.load_runs(runs, self.state)

    def store_test_state(self, obj, testdata=None):

        # Where should I store?
        if testdata is None:
            testdata = obj.testdata

        try:
            os.makedirs(self.statedir)
        except:
            pass

        metadata = json.dumps(
            {
                "name": self.name,
                "runid": self.state.runid,
                "notes": self.state.notes,
                "performance": self.state.performance_summary,
                "stats": self.state.stats,
                "kpi": self.state.kpi,
                "start_time": self.state.start_time,
                "end_time": self.state.end_time,
                "pid": self.state.pid,
                "frames": self.state.state,
                "expectations": self.state.expectations,
                "accesses": self.accessed_files,
                "dataset_accesses": self.state.dataset_accesses,
                "versionmap": self.versionmap,
            },
            cls=SafeEncoder,
            indent=4,
        )

        path = os.path.join(self.statedir, "metadata.json")
        with open(path, "w") as fd:
            fd.write(metadata)

        frames = self.state.get_frame_list()
        for frame in frames:
            detail = self.state.get_frame(frame)

            # Parameters for export
            export = {
                "index": False,
            }

            for p in detail["params"]:
                if p.get("type", None) == "export":
                    export.update(p.get("params", {}))

            df = detail["df"]
            if df is None:
                continue

            frametype = detail["frametype"]
            if frametype in ["pandas", "spark"]:
                path = os.path.join(self.statedir, frame + ".csv")
                df.to_csv(path, **export)
            elif frametype in ["dict"]:
                path = os.path.join(self.statedir, frame + ".json")
                with open(path, "w") as fd:
                    fd.write(json.dumps(df, indent=4))

        return self.statedir

    def get_state(self):
        return self.state

    def add_notification_targets(self, targets):
        if "email" in targets:
            self.notifications["email"].extend(targets["email"])

    def get_credentials(self, name):

        siteconf = self.siteconf

        if ("credentials" not in siteconf) or (name not in siteconf["credentials"]):
            raise Exception("missing credentials")

        return siteconf["credentials"][name]

    def get_versionmap(self):
        return copy.copy(self.versionmap)


class MockPipeline:
    """
    Simulate a pipeline
    """

    def execute(self, transformcls, testdata, save_state=False, context={}):

        # => First create a mock pipeline.
        config = MockConfig(context)

        # => Configure it...
        config_params = testdata.get("config", {})
        config.configure(config_params)

        # => Load test state
        config.load_test_state(testdata)

        # => Get the state object..
        state = config.get_state()

        # => Instantiate the transform...
        transform = transformcls(config=config)

        # => Configure
        conf = testdata.get("conf", {})
        transform.configure(conf)

        # => Validate the configuration
        transform.validate("conf", state)
        transform.validate("args", state)

        # => Now call the process
        transform.process(state)

        # => Now validate the output state
        transform.validate("results", state)

        # => Store the final resultant state
        if save_state:
            config.store_test_state(transform, testdata)



import os
import resource
import psutil
import time
import logging

logger = logging.getLogger("app")


def set_limits_memory(params={}):

    # Default fraction of max memory to be used
    fraction = params.get("memory", 0.8)

    # Get available virtual memory
    vmem = psutil.virtual_memory()

    max_vmem = int(fraction * vmem.available)

    # use AS as alternative to VMEM if the attribute isn't defined.
    # http://stackoverflow.com/a/30269998/5731870
    if hasattr(resource, "RLIMIT_VMEM"):
        resource.setrlimit(resource.RLIMIT_VMEM, (max_vmem, max_vmem))
    elif hasattr(resource, "RLIMIT_AS"):
        resource.setrlimit(resource.RLIMIT_AS, (max_vmem, max_vmem))

    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    soft = round(soft / (1.0 * 1024 * 1024))
    hard = round(hard / (1.0 * 1024 * 1024))

    return "Memory: Soft ({}MB) Hard ({}MB)".format(soft, hard)


def set_limits(params={}):
    msg = set_limits_memory(params)

    logger.debug("Resource Limits set", extra={"transform": "Enricher", "data": msg})


def clear_cache():
    """
    https://www.thegeekdiary.com/how-to-clear-the-buffer-pagecache-disk-cache-under-linux/
    """
    os.system("sudo /bin/bash -c 'echo 1 > /proc/sys/vm/drop_caches'")


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG)
    set_limits({"memory": 0.8})



from enrichsdk.lib.customer import find_usecase

def get_spec():

    return {
        "name": "Catalog",
        "description": "Catalog",
        "usecase": find_usecase(__file__),
        "icon": "grid-icon-25"
    }



"""
Policy firewall for the LLM Agent
"""
import os
import sys
import json
import traceback
import logging
from collections import Counter

logger = logging.getLogger(__file__)

class PolicyBase:

    def __init__(self, *args, **kwargs):
        self.name = "PolicyBase"
        self.description = "Dont use this. Only only the subclass"

    def evaluate(self, text):
        raise Exception("Implement the policy in baseclass")

class DummyPolicy(PolicyBase):
    """
    Do nothing policy
    """

    def __init__(self, *args, **kwargs):
        self.name = "DummyPolicy"
        self.description = "Check for nothing. Always returns success"

    def evaluate(self, text):
        return {
            "status": "success",
            "message": "Dummy policy"
        }

class DuhPolicy(PolicyBase):
    """
    Do nothing policy
    """

    def __init__(self, *args, **kwargs):
        self.name = "DuhPolicy"
        self.description = "Check for duh in evaluation text"

    def evaluate(self, text):
        return {
            "status": "success" if "duh" in text.lower() else "failure",
            "message": "Checks for Duh in the text"
        }


class PolicyManagerBase:
    """

    This has a collection of policysets each of which is a collection of
    policies.

    """
    def __init__(self):
        self.policysets = {}
        self.name = "PolicyManagerBase"

    def add_policy(self, policyset, name, policy):
        """
        """

        if policyset not in self.policysets:
            self.policysets[policyset] = {}

        if name in self.policysets[policyset]:
            logger.warning(f"Policy {name} already present in {policyset} policyset. Updating")

        self.policysets[policyset][name] = policy

    def evaluate(self, name, text):

        if name not in self.policysets:
            return {
                "status": "failure",
                "message": f"Unknown policyset {name}"

            }

        policymap = self.policysets[name]

        results = []
        for name, policy in policymap.items():
            try:
                result = policy.evaluate(text)
                if not isinstance(result, dict):
                    result = {
                        "status": "failure",
                        "message": "Invalid result from verifier",
                        "details": result
                    }
                elif 'status' not in result:
                    result['status'] = 'failure'
                    result['message'] = "Result status is unknown"
            except Exception as e:
                result = {
                    "status": "failure",
                    "message": "Invalid result from verifier",
                    "details":  traceback.format_exc()
                }
            results.append(result)

        if len(results) == 0 and False:
            return {
                "status": "failure",
                "message": "No policies found to evaluate"
            }

        statuses = Counter([r['status'] for r in results])

        if 'failure' in statuses:
            status = 'failure'
            message = "Failed with the following statues: "
            for k, c in statuses.items():
                message += f"{k} ({c}) "
        else:
            status = 'success'
            message = 'Passed all policy checks'

        return {
            'status': status,
            'message': message,
            'details': results
        }

class SimplePolicyManager(PolicyManagerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SimplePolicyManager"

if __name__ == "__main__":

    p = DummyPolicy()
    mgr = SimplePolicyManager()

    mgr.add_policy("default", "dummy", p)
    result = mgr.evaluate("default", "Check this dummy text")
    print(json.dumps(result, indent=4))

    mgr.add_policy("default", "duh", DuhPolicy())
    result = mgr.evaluate("default", "Check this dummy text")
    print(json.dumps(result, indent=4))

    result = mgr.evaluate("default", "Check this duh text")
    print(json.dumps(result, indent=4))



import json
import sqlite3
import logging
import traceback
import re
import time
from sqllineage.runner import LineageRunner

from dateutil import parser as dateparser
from datetime import datetime, timedelta
from collections import defaultdict

from google.cloud import logging_v2 as gcplogging
from google.auth import load_credentials_from_file

logger = logging.getLogger("app")

#######################################
class StorageBase(object):
    def __init__(self, config, *args, **kwargs):
        self.config = config


class SQLiteStorage(StorageBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SQLite"

        if "db" not in self.config:
            raise Exception("db should be specified")

        self.initialize()

    def initialize(self):

        con = sqlite3.connect(self.config["db"])
        cursor = con.cursor()
        sql = """
CREATE TABLE IF NOT EXISTS raw (
     id INTEGER PRIMARY KEY AUTOINCREMENT,
     loggroup TEXT NOT NULL,
     logstream TEXT NOT NULL,
     lastevent BIGINT,
     timestamp BIGINT,
     message TEXT NOT NULL
)
        """

        cursor.execute(sql)

        sql = """
CREATE TABLE IF NOT EXISTS processed (
     id INTEGER PRIMARY KEY AUTOINCREMENT,
     loggroup text NOT NULL,
     logstream TEXT NOT NULL,
     lastevent BIGINT,
     timestamp DATETIME,
     level VARCHAR(256),
     path VARCHAR(256),
     lineno VARCHAR(256),
     method VARCHAR(128),
     message TEXT NOT NULL,
     metadata TEXT
)
        """
        cursor.execute(sql)
        con.close()

    def check(self, table, group, stream, lastevent):

        con = sqlite3.connect(self.config["db"])

        sql = f"""\
SELECT count(1)
FROM {table}
WHERE ((loggroup = '{group}') AND (logstream = '{stream}') and (lastevent = '{lastevent}'))
        """

        cursor = con.cursor()
        cursor.execute(sql)
        rows = cur.fetchall()
        count = rows[0][0]
        con.close()

        return count > 0

    def get_lastevent(self, table, group, stream):

        con = sqlite3.connect(self.config["db"])

        sql = f"""\
select DISTINCT lastevent
FROM {table}
WHERE ((loggroup = '{group}') AND (logstream = '{stream}'))
        """

        cursor = con.cursor()
        cursor.execute(sql)
        rows = cursor.fetchall()
        lastevent = rows[0][0] if len(rows) > 0 else None
        con.close()
        return lastevent

    def clear(self, table, group, stream):

        con = sqlite3.connect(self.config["db"])

        sql = f"""\
DELETE
FROM {table}
WHERE ((loggroup = '{group}') AND (logstream = '{stream}'))
        """
        cursor = con.cursor()
        cursor.execute(sql)
        con.commit()
        con.close()
        return

    def insert(self, table, group, stream, lastevent, events):

        con = sqlite3.connect(self.config["db"])

        # => Construct the values...
        cols = ["timestamp", "level", "path", "lineno", "method", "message", "metadata"]
        values = []
        for e in events:
            v = [group, stream, lastevent]
            for col in cols:
                v.append(e[col])
            values.append(v)

        # construct sql
        sql = f"""\
INSERT INTO {table}( loggroup, logstream, lastevent, timestamp, level, path, lineno, method, message, metadata)
VALUES (?, ?, ?, ?,?, ?, ?, ?,?,?)
        """
        cursor = con.cursor()
        cursor.executemany(sql, values)
        con.commit()
        con.close()


######################################
#
######################################
class TaggerBase(object):
    def __init__(self, config, *args, **kwargs):
        self.config = config


class RegexTagger(TaggerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "RegexTagger"

    def tag(self, event):

        rules = self.config["rules"]

        message = event["message"]
        tags = []
        for tagname, patterns in rules.items():
            for p in patterns:
                match = re.search(p, message, re.IGNORECASE)
                if match is not None:
                    tags.append(tagname)
                    break
        if len(tags) == 0:
            tags.append("no-tag")
        return tags


class MetadataBase(object):
    def __init__(self, config, *args, **kwargs):
        self.config = config

    def process(self, event):
        return {}


class DBTLineageMetadata(MetadataBase):
    def process(self, event):

        message = event["message"]

        if not ((re.search(r"On \S+:", message)) and (r'"app": "dbt"' in message)):
            return {}

        lines = message.split("\n")

        # First compute the SQL
        sql = ""
        join = False
        for idx, m in enumerate(lines):
            if (re.match(r"^On \S+:", m)) and ('"app": "dbt"' in m):
                join = True
                continue
            if join:
                sql += "\n" + m

        update = {}
        if len(sql) == 0:
            return update

        try:
            result = LineageRunner(sql, verbose=True)
            update = {
                "lineage": {
                    "status": "success",
                    "source_tables": [str(t) for t in result.source_tables],
                    "target_tables": [str(t) for t in result.target_tables],
                    "intermediate_tables": [str(t) for t in result.intermediate_tables],
                    "column_lineage": result.get_column_lineage(),
                }
            }

        except:
            update = {
                "lineage": {"status": "failure", "exception": traceback.format_exc()}
            }

        return update


class ProcessorBase(object):
    pass


class CloudwatchProcessor(ProcessorBase):
    def __init__(self, *args, **kwargs):
        self.name = "Cloudwatch"

    def get_streams(self, spec):

        client = spec["args"]["client"]
        log_group = spec["log_group"]

        # What is the cutoff
        cutoff = dateparser.parse(spec.get("cutoff", "1970-01-01"))
        cutoff = (cutoff - datetime(1970, 1, 1)).total_seconds()

        msg = ""
        allstreams = []
        next_token = None
        iteration = 1
        max_iterations = 100
        while iteration < max_iterations:
            try:
                params = {"nextToken": next_token} if next_token is not None else {}
                response = client.describe_log_streams(
                    logGroupName=log_group,
                    orderBy="LastEventTime",
                    descending=True,
                    limit=50,
                    **params,
                )

                # Get all the streams
                streams = response["logStreams"]
                numstreams = len(streams)

                # Now filter out older ones
                streams = [
                    s for s in streams if (s["firstEventTimestamp"] / 1000) >= cutoff
                ]
                netstreams = len(streams)
                msg += f"[{iteration}] {numstreams} -> {netstreams}\n"

                if netstreams == 0:
                    break

                allstreams.extend(streams)
                next_token = response["nextToken"]

                # Some records have been filtered, i.e., we
                # cross the cutoff
                if netstreams < numstreams:
                    break

                iteration += 1
            except:
                logger.exception(
                    f"Failed {log_group}", extra={"transform": self.name, "data": msg}
                )
                raise

        msg += f"[Completed] {len(allstreams)}"
        logger.debug(
            f"[{log_group}] {len(allstreams)} streams",
            extra={"transform": self.name, "data": msg},
        )

        return allstreams

    def get_logs(self, spec, stream):

        """
        {
             "logStreamName": "ecs/trip-recommender-api/3c4f22c7e0444331973bc040ce577265",
             "creationTime": 1637033546947,
             "firstEventTimestamp": 1637033588852,
             "lastEventTimestamp": 1637033589298,
             "lastIngestionTime": 1637033592949,
             "uploadSequenceToken": "49605172774206290057587551428896880961727365107853296162",
             "arn": "arn:aws:logs:eu-west-1:441203537012:log-group:/ecs/trip-recommender-api:log-stream:ecs/trip-recommender-api/3c4f22c7e0444331973bc040ce577265",
             "storedBytes": 0
         },
        """

        client = spec["args"]["client"]
        db = spec["args"]["db"]

        # What is the cutoff
        cutoff = dateparser.parse(spec.get("cutoff", "1970-01-01"))
        cutoff = (cutoff - datetime(1970, 1, 1)).total_seconds()

        specname = spec["name"]
        log_group = spec["log_group"]
        log_stream = stream["logStreamName"]
        created = stream["creationTime"]
        start_time = max(stream["firstEventTimestamp"], cutoff)
        lastevent_timestamp = stream["lastEventTimestamp"]
        name = stream["logStreamName"]
        force = spec.get("force", False)

        # Get existing last event
        db_lastevent = db.get_lastevent("processed", log_group, log_stream)
        msg = f"[DB] Last event: {db_lastevent} {type(db_lastevent)}\n"
        msg += (
            f"[Stream] Last event: {lastevent_timestamp} {type(lastevent_timestamp)} ("
            + ("Match" if (db_lastevent == lastevent_timestamp) else "No match")
            + ")\n"
        )
        msg += f"[Force] {force}\n"

        # Exists and valid..
        if (
            (db_lastevent is not None)
            and (db_lastevent == lastevent_timestamp)
            and (not force)
        ):
            msg += "Skipping. Updated"
            logger.debug(
                f"[{log_stream[-32:]}] Skipping",
                extra={"transform": self.name, "data": msg},
            )
            return None

        # Clear all previous events
        if db_lastevent is not None:
            msg += "Clearing db\n"
            db.clear("processed", log_group, log_stream)

        allevents = []
        iteration = 1
        next_token = None
        max_iterations = 5 if spec.get("test", False) else 10000
        while iteration < max_iterations:
            try:
                params = {"nextToken": next_token} if next_token is not None else {}
                response = client.get_log_events(
                    logGroupName=log_group,
                    logStreamName=log_stream,
                    startTime=start_time,
                    endTime=lastevent_timestamp,
                    startFromHead=False,
                    limit=2000,
                    **params,
                )

                events = response["events"]
                allevents.extend(events)
                if len(events) == 0:
                    break

                iteration += 1
                next_token = response["nextBackwardToken"]

                msg += (
                    f"Iteration: {iteration} new {len(events)} Total {len(allevents)}\n"
                )

            except:
                logger.exception(
                    f"Failed {log_stream[-32:]}",
                    extra={
                        "transform": self.name,
                    },
                )
                raise

        logger.debug(
            f"[{log_stream[-32:]}] {len(allevents)} events",
            extra={"transform": self.name, "data": msg},
        )

        return allevents

    def collapse_events(self, spec, stream, allevents):
        """
        Take the raw events and process them...
        """

        db = spec["args"]["db"]
        specname = spec["name"]
        regexps = spec["regexps"]
        if isinstance(regexps, str):
            regexps = [regexps]
        log_group = spec["log_group"]
        log_stream = stream["logStreamName"]

        created = stream["creationTime"]
        lastevent_timestamp = stream["lastEventTimestamp"]

        try:
            stats = defaultdict(int)
            cleaned = []
            lastevent = None
            for event in allevents:

                stats["total"] += 1
                timestamp = event["timestamp"]
                dt = datetime.utcfromtimestamp(timestamp / 1000)
                dt = dt.replace(microsecond=(timestamp % 1000) * 1000)

                message = event["message"]

                # 2022-01-21 11:04:19,106 loglevel=INFO   logger=src.matching load_s3_pickle() L45
                match = None
                for regexp in regexps:
                    match = re.match(regexp, message)
                    if match is not None:
                        break

                parsedevent = None
                if match is not None:
                    stats["parsed"] += 1
                    parsedevent = match.groupdict()

                    # Use the log timestamp consistently. This differs from the log timestamp.
                    parsedevent["timestamp"] = dt.isoformat()
                    for col in ["path", "method", "lineno"]:
                        if col not in parsedevent:
                            parsedevent[col] = None

                defaultevent = {
                    "level": None,
                    "timestamp": dt.isoformat(),
                    "path": None,
                    "method": None,
                    "lineno": None,
                    "message": message,
                }

                # print("last event", lastevent)

                if parsedevent is not None:
                    if lastevent is not None:
                        cleaned.append(lastevent)
                    lastevent = parsedevent
                    stats["final"] += 1
                    continue
                elif lastevent is not None:
                    lastevent["message"] += "\n" + message
                    stats["collapsed"] += 1
                else:
                    stats["final"] += 1
                    lastevent = defaultevent

            if lastevent is not None:
                stats["final"] += 1
                cleaned.append(lastevent)

            filtered = []
            tagger = spec["args"]["tagger"]
            lastevent = None
            for event in cleaned:

                tags = tagger.tag(event)

                if "ignore" in tags:
                    stats["ignored"] += 1
                    continue

                event["metadata"] = {"tags": tags}

                # 2022-01-24T12:00:27.685000
                # Merge if the tags the same, within a min of the previous event
                if (
                    (lastevent is not None)
                    and (set(tags) == set(lastevent["metadata"]["tags"]))
                    and (event["timestamp"][:16] == lastevent["timestamp"][:16])
                ):
                    lastevent["message"] = (
                        event["message"] + "\n" + lastevent["message"]
                    )
                    stats["merged"] += 1
                    continue

                lastevent = event
                filtered.append(event)

            stats["final"] -= stats.get("ignored", 0)
            stats["final"] -= stats.get("merged", 0)

            for event in filtered:
                event["metadata"] = json.dumps(event["metadata"])

            db.insert("processed", log_group, log_stream, lastevent_timestamp, filtered)
            logger.debug(
                f"[{log_stream[-32:]}] processed",
                extra={"transform": self.name, "data": json.dumps(stats, indent=4)},
            )
        except:
            logger.exception(
                f"[{log_stream[-32:]}] failed",
                extra={"transform": self.name, "data": json.dumps(stats, indent=4)},
            )
            raise

    def run(self, spec):

        logger.debug(f"Processing {spec['name']}", extra={"data": str(spec)})

        streams = self.get_streams(spec)
        for stream in streams:
            try:
                allevents = self.get_logs(spec, stream)
                if (allevents is None) or (
                    isinstance(allevents, list) and len(allevents) == 0
                ):
                    continue

                self.collapse_events(spec, stream, allevents)

            except:
                logger.exception(f"Failed {stream['logStreamName']}")

        logger.debug(f"Completed {spec['name']}", extra={"data": str(spec)})


class S3DBTProcessor(ProcessorBase):
    def __init__(self, *args, **kwargs):
        self.name = "S3DBTProcessor"

    def process_one_file(self, spec, loggroup, f):

        s3 = spec["args"]["s3"]
        counts = defaultdict(int)
        regexps = spec["regexps"]

        # => Get the lines
        with s3.open(f, "r") as fd:
            lines = fd.readlines()
            lines = [l.rstrip() for l in lines]

        allstreams = {}
        logstream = None
        current_entry = None
        for line in lines:

            try:

                counts["lines"] += 1
                # First match the line and to see any regexp works
                for regexp in regexps:
                    match = re.match(regexp, line)
                    if match is not None:
                        break

                # Extract matched data. could be
                data = None if (match is None) else match.groupdict()

                if (logstream is None) or (
                    (data is not None) and ("streamname" in data)
                ):

                    # We havent seen the first line for the run
                    if match is None:
                        counts["nostream_nomatch"] += 1
                        continue

                    if "streamname" not in data:
                        counts["nostream_match_nodata"] += 1
                        continue

                    logstream = data["streamname"]
                    start_timestamp = dateparser.parse(data["timestamp"])
                    lastevent = (
                        time.mktime(start_timestamp.timetuple()) * 1e3
                        + start_timestamp.microsecond / 1e3
                    )

                    counts["stream_new"] += 1

                    # Now create a basic entry
                    allstreams[logstream] = {
                        "lastevent": lastevent,
                        "start_timestamp": start_timestamp,
                        "entries": [],
                    }
                    continue

                # =>We have logstream now. New entry
                start_timestamp = allstreams[logstream]["start_timestamp"]
                entries = allstreams[logstream]["entries"]
                if data is None:
                    if current_entry is not None:
                        current_entry["message"] += "\n" + line
                        counts["stream_merge"] += 1
                    else:
                        counts["stream_nocurrent_entry"] += 1
                    continue

                # New parsed entry with a timestamp. Store the old one
                if current_entry is not None:
                    entries.append(current_entry)
                    current_entry = None

                # Data is not None and logstream exists
                try:
                    # 05:36:37.028275 [info ] [MainThread]: Running with dbt=1.0.1
                    ts = dateparser.parse(data["timestamp"])
                    ts = ts.replace(
                        year=start_timestamp.year,
                        month=start_timestamp.month,
                        day=start_timestamp.day,
                    )
                    if ts < start_timestamp:
                        ts += timedelta(days=1)

                    level = data["level"]
                    message = data["message"]
                    counts["stream_newentry"] += 1
                    current_entry = {
                        "loggroup": loggroup,
                        "logstream": logstream,
                        "lastevent": lastevent,
                        "timestamp": ts.isoformat(),
                        "level": level,
                        "path": None,
                        "lineno": None,
                        "method": None,
                        "message": message,
                    }
                except:
                    traceback.print_exc()
                    counts["stream_invalid_entry"] += 1
                    raise
            except:
                logger.exception(f"{logstream} Unable to read")
                counts["stream_exception"] += 1
                raise

        if current_entry is not None and logstream is not None:
            allstreams[logstream]["entries"].append(current_entry)

        return allstreams, counts

    def run(self, spec):

        logger.debug(f"Processing {spec['name']}", extra={"data": str(spec)})

        log_group = spec["name"]
        db = spec["args"]["db"]
        s3 = spec["args"]["s3"]
        filepattern = spec["args"]["filepattern"]
        logstyle = spec["args"].get("logstyle", "default")

        if logstyle != "default":
            raise Exception(f"Unsupported logstyle. Only 'default' supported")

        # => Process all the files...
        files = s3.glob(filepattern)
        for idx, f in enumerate(files):
            allstreams, counts = self.process_one_file(spec, log_group, f)
            msg = f + "\n"
            msg += "Stats: " + json.dumps(counts, indent=4) + "\n"
            for logstream, detail in allstreams.items():
                msg += f"[{logstream}] {len(detail['entries'])}\n"
            logger.debug(f"[{log_group}] [{idx}] Processed", extra={"data": msg})

        # => Now tag and merge them...
        msg = ""
        tagger = spec["args"]["tagger"]
        for logstream in list(allstreams.keys()):

            entries = allstreams[logstream]["entries"]
            stats = defaultdict(int)

            # Tag and merge
            lastevent = None
            filtered = []
            for event in entries:

                stats["total"] += 1
                tags = tagger.tag(event)
                if "ignore" in tags:
                    stats["ignored"] += 1
                    continue
                event["metadata"] = {"tags": tags}

                if lastevent is None:
                    lastevent = event
                    filtered.append(event)
                    stats["new"] += 1
                    continue

                # 2022-01-24T12:00:27.685000
                # Merge if the tags the same, within a min of the previous event
                if (set(tags) == set(lastevent["metadata"]["tags"])) and (
                    event["timestamp"][:16] == lastevent["timestamp"][:16]
                ):
                    lastevent["message"] = (
                        event["message"] + "\n" + lastevent["message"]
                    )
                    stats["merged"] += 1
                    continue

                lastevent = event
                filtered.append(event)
                stats["new"] += 1

            if "metadata" in spec["args"]:
                metadata = spec["args"]["metadata"]
                for entry in filtered:
                    update = metadata.process(entry)
                    if len(update) > 0:
                        stats["nontag_metadata"] += 1
                        entry["metadata"].update(update)

            # Serialize entries...
            for entry in filtered:
                entry["metadata"] = json.dumps(entry["metadata"])

            # Now store it back...
            allstreams[logstream]["filtered"] = filtered
            msg += f"[{logstream}]\n" + json.dumps(stats, indent=4) + "\n"

        logger.debug("Tagged and merged", extra={"data": msg})
        # for event in filtered:
        #    print("-----")
        #   print(event['timestamp'])
        #  print(event['message'])

        # =>
        db = spec["args"]["db"]
        for log_stream, details in allstreams.items():
            filtered = details["filtered"]
            lastevent = details["lastevent"]
            db.clear("processed", log_group, log_stream)
            db.insert("processed", log_group, log_stream, lastevent, filtered)

        logger.debug(f"Completed {log_group}")


class StackDriverProcessor():
    """
    Read interface to GCP Stack driver. It has limitations in terms of
    being able to read one GCP log stream alone (due to the use of
    google.cloud.logging library instead of google service API.
    """
    def __init__(self, *args, **kwargs):
        self.name = "Stackdriver"

    def list_entries(self, spec, stream):
        """
        Args:
          spec (object): Dict containing 'cred' (credentials path)
          stream (object): Dict containing name (humanized name), logname (GCP log name e.g., projects/<project_id>/logs/<logName>), start

        Returns:
          iterator (object):  One entry at a time

        Sample code::

            spec = {
                'cred': os.environ['CREDENTIALS_FILE']
            }

            stream = {
                'name': "syslog",
                "logname": "projects/scribble-internal-1/logs/syslog",
                "start_time": "2022-05-25T07:50:00Z",
                "end_time": "2022-05-25T08:50:00Z",
            }

            for entry in driver.list_entries(spec, stream):
                print(entry)


        """

        cred = spec['cred']
        sdclient = gcplogging.Client(credentials=cred)

        name = stream['name']
        logname = stream['logname']

        sdlogger = sdclient.logger(name)

        start_time = stream.get("start_time", "1970-01-01T00:00:00Z")
        end_time = stream.get('end_time', datetime.utcnow().isoformat())

        filtercmd = f"""\
logName="{logname}"
timestamp >= "{start_time}"
timestamp <= "{end_time}"
"""

        if 'extra' in stream:
            filtercmd += str(stream['extra'])

        logger.debug(f"Accessing stackdriver: {name}",
                     extra={
                         'data': filtercmd
                     })

        for entry in sdlogger.list_entries(filter_=filtercmd):
            yield entry




r"""
This developer kit is meant for advanced users of the `Enrich Full-Stack Feature Store`. This includes command line scripts to create and manage Enrich modules and APIs to write new modules on the platform:

<pre>
 _____            _      _       ____  ____  _  __
| ____|_ __  _ __(_) ___| |__   / ___||  _ \| |/ /
|  _| | '_ \| '__| |/ __| '_ \  \___ \| | | | ' /
| |___| | | | |  | | (__| | | |  ___) | |_| | . \
|_____|_| |_|_|  |_|\___|_| |_| |____/|____/|_|\_\

</pre>

"""
import os

VERSION = "5.0.7"


def _get_git_revision(path):
    revision_file = os.path.join(path, "refs", "heads", "master")
    if os.path.exists(revision_file):
        with open(revision_file) as fh:
            return fh.read().strip()[:7]


def get_revision():
    #
    #:returns: Revision number of this branch/checkout, if available. None if
    #    no revision number can be determined.
    #
    package_dir = os.path.dirname(__file__)
    checkout_dir = os.path.normpath(os.path.join(package_dir, os.pardir))
    path = os.path.join(checkout_dir, ".git")
    if os.path.exists(path):
        return _get_git_revision(path)


def get_version():
    """
    Revision number of this branch/checkout, if available. None if
    no revision number can be determined.
    """

    base = VERSION
    if __build__:
        base = "%s (%s)" % (base, __build__)
    return base


__build__ = get_revision()
__version__ = VERSION

from .core import *
from .package import *
from .lib import *
from .commands import *
from .tasks import *
from .services import *
from .datasets import *
from .quality import *
from . import featurestore, realtime, contrib, utils, notebook, feature_compute
